/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/


/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

package ojb.broker;
import java.util.List;
import ojb.broker.metadata.DescriptorRepository;
import ojb.broker.metadata.ClassDescriptor;
import java.io.*;
import java.util.zip.*;

/**
 *
 * represents the identity of an object.
 *
 * @author: thma
 *
 */
public class Identity implements java.io.Serializable
{

	/**
	 * the Class of the identified object
	 */
	private Class objectsClass;

	/**
	 * The ordered list of primary key values maintaining the objects identity in the underlying RDBMS
	 */
	private Object[] pkValues;
	private transient String stringRepresentation = null;

	/**
	 *
	 * empty constructor is private.
	 *
	 */
	private Identity()
	{
		super();
	}

	/**
	 * creates an Identity from an object. Reads the objects primary key values.
	 * @param objectToIdentitify java.lang.Object
	 */
	public Identity(Object objectToIdentitify)
	{
		try
		{
			ClassDescriptor mif =
				DescriptorRepository.getInstance().getDescriptorFor(objectToIdentitify.getClass());
			this.objectsClass = objectToIdentitify.getClass();
			this.pkValues = mif.getKeyValues(objectToIdentitify);
		}
		catch (ojb.broker.ClassNotPersistenceCapableException ex)
		{
			throw new org.odmg.ClassNotPersistenceCapableException(ex.getMessage());
		}
	}

	/**
	 * factory method that returns an Identity object from the serialized Identity in anArray.
	 */
	public static Identity fromByteArray(byte[] anArray)
	{
	    // reverse of the serialize() algorithm:
	    // read from byte[] with a ByteArrayInputStream, decompress with
	    // a GZIPInputStream and then deserialize by reading from the ObjectInputStream
		try
		{
			ByteArrayInputStream bais = new ByteArrayInputStream(anArray);
            GZIPInputStream gis = new GZIPInputStream(bais);
			ObjectInputStream ois = new ObjectInputStream(gis);
			Identity result = (Identity) ois.readObject();
			ois.close();
			gis.close();
			bais.close();
			return result;
		}
		catch (Exception ex)
		{
			return null;
		}
	}

	/**
	 *
	 * return the Class of the real subject
	 *
	 * @return java.lang.Class
	 *
	 */
	public java.lang.Class getObjectsClass()
	{
		return objectsClass;
	}

	/**
	 *
	 * return a serialized Identity as byte[]
	 *
	 * @return byte[]
	 *
	 */
	public byte[] serialize()
	{
	    // Identity is serialized and written to an ObjectOutputStream
	    // This ObjectOutputstream is compressed by a GZIPOutputStream
	    // and finally written to a ByteArrayOutputStream.
	    // the resulting byte[] is returned
		try
		{
			ByteArrayOutputStream bao = new ByteArrayOutputStream();
			GZIPOutputStream gos = new GZIPOutputStream(bao);
			ObjectOutputStream oos = new ObjectOutputStream(gos);
			oos.writeObject(this);
			oos.close();
			gos.close();
			bao.close();
			byte[] result = bao.toByteArray();
			return result;
		}
		catch (Exception ignored)
		{
			return null;
		}
	}

	/**
	 *
	 * return a String representation.
	 *
	 * @return java.lang.String
	 *
	 */
	public String toString()
	{
		if (stringRepresentation == null)
		{
			StringBuffer buf = new StringBuffer();
			for (int i = 0; i < pkValues.length; i++)
			{
				buf.append((buf.length() == 0) ? "{" : ",");
				buf.append(pkValues[i]);
			}
			buf.append("}");
			stringRepresentation = objectsClass.getName() + buf.toString();
		}
		return stringRepresentation;
	}

	/**
	 *
	 * creates an Identity from a class and the objects primary key values.
	 *
	 * used for the definition of proxies.
	 *
	 * @param aClass java.lang.Class
	 *
	 * @param pkValues List
	 *
	 */
	public Identity(Class aClass, Object[] pkValues)
	{
		this.objectsClass = aClass;
		this.pkValues = pkValues;
	}

	/**
	 *
	 * return the list of Primary Key Values of the real subject
	 *
	 * @return Object[]
	 *
	 */
	public Object[] getPrimaryKeyValues()
	{
		return pkValues;
	}

}
