/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.broker;
import java.util.Collection;
import java.util.Iterator;
import java.util.Enumeration;
import org.odmg.TransactionAbortedException;
import org.odmg.TransactionNotInProgressException;
import java.util.List;
import ojb.broker.query.Query;
import org.odmg.TransactionInProgressException;

/**
 *
 * PersistenceBroker declares a protocol for persisting arbitrary objects.
 *
 * A typical implementation might wrap an RDBMS access layer.
 *
 * @see PersistenceBrokerClient
 *
 *
 *
 * @author <a href="mailto:thomas.mahler@itellium.com">Thomas Mahler<a>
 *
 */
public interface PersistenceBroker
{

	/**
	 * 
	 * deletes the objects obj concrete representation in the underlying persistence system.
	 * 
	 * E.G. by DELETE FROM ... WHERE ... in an RDBMS
	 * 
	 */
	public void delete(Object obj) throws PersistenceBrokerException;

	/**
	 * 
	 * make object obj persistent in the underlying persistence system.
	 * 
	 * E.G. by INSERT INTO ... or UPDATE ...  in an RDBMS
	 * 
	 */
	public void store(Object obj) throws PersistenceBrokerException;

	/**
	 * 
	 * Abort and close the transaction.
	 * 
	 * Calling abort abandons all persistent object modifications and releases the
	 * 
	 * associated locks.
	 * 
	 * If transaction is not in progress a TransactionNotInProgressException is thrown
	 * 
	 */
	public void abortTransaction() throws TransactionNotInProgressException;

	/**
	 * 
	 * begin a transaction against the underlying RDBMS.
	 * 
	 * Calling <code>beginTransaction</code> multiple times,
	 * 
	 * without an intervening call to <code>commitTransaction</code> or <code>abortTransaction</code>,
	 * 
	 * causes the exception <code>TransactionInProgressException</code> to be thrown
	 * 
	 * on the second and subsequent calls.
	 * 
	 */
	public void beginTransaction() 
			throws TransactionInProgressException, org.odmg.TransactionAbortedException;

	/**
	 * 
	 * Commit and close the transaction.
	 * 
	 * Calling <code>commit</code> commits to the database all
	 * 
	 * UPDATE, INSERT and DELETE statements called within the transaction and
	 * 
	 * releases any locks held by the transaction.
	 * 
	 * If beginTransaction() has not been called before a
	 * 
	 * TransactionNotInProgressException exception is thrown.
	 * 
	 * If the transaction cannot be commited a TransactionAbortedException exception is thrown.
	 * 
	 */
	public void commitTransaction() 
			throws TransactionNotInProgressException, TransactionAbortedException;

	/**
	 * 
	 * retrieve a collection of itemClass Objects matching the Query query.
	 * 
	 * if the Query has no criteria no WHERE-clause is generated, i.e. ALL table rows are selected.
	 * 
	 */
	public Collection getCollectionByQuery(Query query) throws PersistenceBrokerException;

    /**
	 *
	 * retrieve a userdefined Collection that implements the interface Manageable collection
	 * that contains all Objects matching the Query query.
	 *
	 * If query has no criteria no WHERE-clause is generated, i.e. ALL table rows are selected.
	 *
	 */

    public ManageableCollection getCollectionByQuery(Class collectionClass, Query query)
        throws PersistenceBrokerException;


	/**
	 * 
	 * returns an Iterator that iterates Objects of class c if calling the .next()
	 * 
	 * method. The Elements returned come from a SELECT ... WHERE Statement
	 * 
	 * that is defined by the Query query.
	 * 
	 * If itemProxy is null, no proxies are used.
	 * 
	 * NOT YET AWARE OF EXTENTS !
	 * 
	 */
	public Iterator getIteratorByQuery(Query query) throws PersistenceBrokerException;

	/**
	 * retrieve an Object by its Identity. Application Developers are encouraged to use getObjectByQuery().
	 * This method is mainly used for internal performant handling of
	 * materialization by OID (e.g. in Proxies)
	 */
	public Object getObjectByIdentity(Identity id) throws PersistenceBrokerException;

	/**
	 * retrieve an Object by query.
	 * I.e perform a SELECT ... FROM ... WHERE ...  in an RDBMS
	 */
	public Object getObjectByQuery(Query query) throws PersistenceBrokerException;

	/**
	 * returns an Enumeration of PrimaryKey Objects for objects of class DataClass.
	 * The Elements returned come from a SELECT ... WHERE Statement
	 * that is defined by the fields and their coresponding values of vecFields
	 * and vecValues.
	 * Useful for EJB Finder Methods...
	 * NOT YET AWARE OF EXTENTS !
	 * @param PrimaryKeyClass the pk class for the searched objects
	 * @param query the query
	 */
	public Enumeration getPKEnumerationByQuery(Class PrimaryKeyClass, 
											   Query query) throws PersistenceBrokerException;   

	/**
	 * 
	 * makes object obj persistent in the underlying persistence system.
	 * 
	 * E.G. by INSERT INTO ... or UPDATE ...  in an RDBMS.
	 * 
	 * The ObjectModification parameter can be used to generate optimized SQL code
	 * 
	 * (decide whether insert or update is needed. And for updates only generate code for modified columns)
	 * 
	 */
	public void store(Object obj, 
					  ObjectModification modification) throws PersistenceBrokerException;  

	/**
	 * returns a unique int for class extent and field attribute.
	 *
	 * the returned uid is unique accross all tables in the extent of class extent.
	 *
	 */
    public int getUniqueId(Class extent, String attribute) throws PersistenceBrokerException;

}