/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.broker;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Vector;
import java.util.Collection;
import java.util.Iterator;
import java.lang.reflect.Field;
import ojb.broker.metadata.*;
import ojb.broker.accesslayer.*;
import java.io.IOException;
import ojb.broker.query.*;
import java.util.List;
import java.net.MalformedURLException;
import org.xml.sax.SAXException;
import java.util.Enumeration;
import javax.xml.parsers.ParserConfigurationException;
import java.lang.reflect.Array;
import java.sql.Connection;

/**
 * The PersistenceBrokerClient is an implementation of the PersistenceBroker
 * Interface that specifies a persistence mechanism for generic objects.
 * This Concrete implementation provides an object relational mapping
 * and allows to store and retrieve arbitrary objects in/from relational
 * databases accessed by JDBC.
 *
 * @see ojb.examples.broker.App for a sample application
 * @author <a href="mailto:thomas.mahler@itellium.com">Thomas Mahler<a>
 */
public class PersistenceBrokerImpl implements PersistenceBroker
{


    public PersistenceBrokerImpl(DescriptorRepository repository)
	{
		m_DescriptorRepository = repository;
		m_ConnectionManager = new ConnectionManager(this);
		m_DbAccess = new JdbcAccess(this);
		m_SequenceManager = new SequenceManager(this);
	}

	/**
	 * debugging flag, false by default = no debugging
	 */
	private static boolean _debug = false;

	/**
	 * set the debugging flag to true to enable debugging
	 */
	public static void setDebug(boolean b)
	{
		_debug = b;
	}

	/**
	 * m_DbAccess is used to do all Jdbc related work: connecting, executing...
	 */
	private JdbcAccess m_DbAccess;

	/**
	 * m_SqlGen is a simple SQL Generator that produces all SQL Queries
	 */
	private SqlGenerator m_SqlGen = SqlGenerator.getInstance();

	/**
	 * m_ObjectCache caches object already loaded from db and protects the uniqueness
	 * of objects. When an object is loaded from the db it is stored in the cache.
	 * Any later lookups with the same pk values will return the cached object
	 * instead of performing another db lookup.
	 */
	private ObjectCache m_ObjectCache = ObjectCache.getInstance();

	/**
	 * holds mapping information for all classes to be treated by PersistenceBroker
	 */
	private DescriptorRepository m_DescriptorRepository = null;

	private ConnectionManager m_ConnectionManager = null;

	private SequenceManager m_SequenceManager = null;

	/**
	 * deletes Object obj's representation in the underlying RDBMS
	 */
	public void delete(Object obj) throws PersistenceBrokerException
	{
		if (_debug)
		{
			System.out.println("delete " + obj);
		}
		if (obj instanceof VirtualProxy)
		{
			VirtualProxy proxy = (VirtualProxy) obj;
			obj = proxy.getRealSubject();
		}
		ClassDescriptor mif = m_DescriptorRepository.getDescriptorFor(obj.getClass());

		// 1. delete dependend objects  first to avoid FK violations
		deleteReferences(obj, mif.getObjectReferenceDescriptors());
		// 2. delete dependend collections
		deleteCollections(obj, mif.getCollectionDescriptors());

        // 3. delete object from directly mapped table
		Identity oid = new Identity(obj);
		m_DbAccess.executeDelete(mif, oid);
		// remove obj from the object cache:
		m_ObjectCache.remove(oid);

	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param obj
	 * @param vecRds
	 * 
	 * @see
	 */
	private void deleteReferences(Object obj, Vector vecRds)
	{
		// get all members of obj that are references and delete them
		Iterator i = vecRds.iterator();
		while (i.hasNext())
		{
			ObjectReferenceDescriptor rds = (ObjectReferenceDescriptor) i.next();
			if (rds.getCascadeDelete())
			{
				try
				{
					Object r = rds.getPersistentField().get(obj);
					delete(r);
				}
				catch (Exception ex)
				{
					ex.printStackTrace();
				}
			}
		}
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param obj
	 * @param vecCds
	 * 
	 * @see
	 */
	private void deleteCollections(Object obj, Vector vecCds)
	{
		// get all members of obj that are collections and
		// delete all their elements
		Iterator i = vecCds.iterator();
		while (i.hasNext())
		{
			CollectionDescriptor cds = (CollectionDescriptor) i.next();
			if (cds.getCascadeDelete())
			{
				try
				{
					Object col = cds.getPersistentField().get(obj);
					if (col != null)
					{
						Iterator colIterator;
						if (col instanceof ManageableCollection)
						{
							colIterator = ((ManageableCollection) col).ojbIterator();
						}
						else if (col instanceof Collection)
						{
							colIterator = ((Collection) col).iterator();
						}
						else if (col.getClass().isArray())
						{
							colIterator = new ArrayIterator(col);
						}
						else
						{
							throw new RuntimeException(col.getClass() 
													   + " can not be managed by OJB, use Array, Collection or ManageableCollection instead !");
						}
						while (colIterator.hasNext())
						{
							delete(colIterator.next());
						}
					}
				}
				catch (Exception ex)
				{
					System.out.println(ex.getMessage());
					ex.printStackTrace();
				}
			}
		}
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param obj
	 * 
	 * @throws PersistenceBrokerException
	 * 
	 * @see
	 */
	public void store(Object obj) throws PersistenceBrokerException
	{
		if (_debug)
		{
			System.out.println("store " + obj);
			// ProxyObjects only have to be updated if their real subjects have been loaded
		}
    	// only do something ig obj != null
    	if (obj != null)
    	{
    		if (obj instanceof VirtualProxy)
    		{
    			VirtualProxy proxy = (VirtualProxy) obj;
    			if (proxy.alreadyMaterialized())
    			{
    				obj = proxy.getRealSubject();
    			}
    			else
    			{
    				return;
    			}
    		}
    		ClassDescriptor mif = m_DescriptorRepository.getDescriptorFor(obj.getClass());
    		Identity oid = new Identity(obj);
    		SimpleObjectModificationImpl mod = new SimpleObjectModificationImpl();
    		// check if update or insert is needed
    		boolean doInsert = false;
    		// lookup cache or db to see whether object needs insert or update
    		if (m_DbAccess.materializeObject(mif, oid) == null)
    		{
    			doInsert = true;
    			// now store it:
    		}
    		store(obj, doInsert);
    	}
    	// if Object == null do nothing
    	else return;
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param obj
	 * @param vecRds
	 * 
	 * @see
	 */
	private void storeReferences(Object obj, Vector vecRds)
	{
		// get all members of obj that are references and store them
		Iterator i = vecRds.iterator();
		while (i.hasNext())
		{
			ObjectReferenceDescriptor rds = (ObjectReferenceDescriptor) i.next();
			if (rds.getCascadeStore())
			{
				try
				{
					Object r = rds.getPersistentField().get(obj);
					store(r);
				}
				catch (Exception ex)
				{
					ex.printStackTrace();
				}
			}
		}
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param obj
	 * @param vecCds
	 * 
	 * @see
	 */
	private void storeCollections(Object obj, Vector vecCds)
	{
		// get all members of obj that are collections and store all their
		// elements
		Iterator i = vecCds.iterator();
		while (i.hasNext())
		{
			CollectionDescriptor cds = (CollectionDescriptor) i.next();
			if (cds.getCascadeStore())
			{
				try
				{
					Object col = cds.getPersistentField().get(obj);
					if (col != null)
					{
						Iterator colIterator;
						if (col instanceof ManageableCollection)
						{
							colIterator = ((ManageableCollection) col).ojbIterator();
						}
						else if (col instanceof Collection)
						{
							colIterator = ((Collection) col).iterator();
						}
						else if (col.getClass().isArray())
						{
							colIterator = new ArrayIterator(col);
						}
						else
						{
							throw new RuntimeException(col.getClass() 
													   + " can not be managed by OJB, use Array, Collection or ManageableCollection instead !");
						}
						while (colIterator.hasNext())
						{
							store(colIterator.next());
						}
					}
				}
				catch (Throwable ex)
				{
					System.out.println(ex.getMessage());
					ex.printStackTrace();
				}
			}
		}
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param newObj
	 * @param mif
	 * 
	 * @see
	 */
	public void retrieveReferences(Object newObj, ClassDescriptor mif)
	{
		Field ReferencedField = null;
		Object refObj = null;
		// iterate over all ReferenceDescriptors to retrieve all associated persistent objects
		Vector rdv = mif.getObjectReferenceDescriptors();
		Iterator i = rdv.iterator();
		try
		{
			while (i.hasNext())
			{
				ObjectReferenceDescriptor rds = (ObjectReferenceDescriptor) i.next();
				if (rds.getCascadeRetrieve())
				{
					refObj = getReferencedObject(newObj, rds, mif);
					ReferencedField = rds.getPersistentField();
					ReferencedField.set(newObj, refObj);
				}
			}
		}
		catch (Exception ex)
		{
			// temporary trace messages
			System.out.println("expected type: " + ReferencedField.getType());
			System.out.println("actual type:   " + refObj.getClass());
			ex.printStackTrace();
		}
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param newObj
	 * @param mif
	 * 
	 * @see
	 */
	public void retrieveCollections(Object newObj, ClassDescriptor mif)
	{
		Vector cdv = mif.getCollectionDescriptors();
		Iterator i = cdv.iterator();
		try
		{
			while (i.hasNext())
			{
				CollectionDescriptor cds = (CollectionDescriptor) i.next();
				if (cds.getCascadeRetrieve())
				{
					// this collection type will be used:
					Class collectionClass = cds.getCollectionClass();
					Field collectionField = cds.getPersistentField();
					Class itemsClass = cds.getItemClass();    // class of items in collection
					// Class itemsProxy = cds.getItemProxyClass(); // class of items proxy
					ClassDescriptor itemsMif = m_DescriptorRepository.getDescriptorFor(itemsClass);
					// the Foreign key attributes of the selected objects...
					FieldDescriptor[] fields = cds.getForeignKeyFieldDescriptors(itemsMif);
					// must match the Objects Primary key values
					Object[] values = mif.getKeyValues(newObj);
					// build and execute query
					Query fkQuery = getForeignKeyQuery(itemsMif, fields, values);
					if (collectionClass == null)
					{
						Collection result = getCollectionByQuery(fkQuery);
						// assign collection to objects attribute
						// if attribute has an array type build an array, else assign collection directly
						if (collectionField.getType().isArray())
						{
							int length = result.size();
							Class itemtype = collectionField.getType().getComponentType();
							Object resultArray = Array.newInstance(itemtype, length);
							for (int j = 0; j < length; j++)
							{
								Array.set(resultArray, j, ((Vector) result).get(j));
							}
							collectionField.set(newObj, resultArray);
						}
						else
						{
							collectionField.set(newObj, result);
						}
					}
					else
					{
						ManageableCollection result = getCollectionByQuery(collectionClass, 
								fkQuery);
						// assign collection to objects attribute
						collectionField.set(newObj, result);
					}
				}
			}
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
		}
	}

	/**
	 * 
	 * retrieves an Object reference.
	 * 
	 * If there is a Proxy-class is defined in the ReferenceDescriptor, a Proxy-object is returned.
	 * 
	 * If no Proxy-class is defined, a getObjectByIdentity(...) lookup is performed.
	 * 
	 */
	private Object getReferencedObject(Object obj, ObjectReferenceDescriptor rds, 
									   ClassDescriptor mif) throws PersistenceBrokerException
	{
		Class referencedClass = rds.getItemClass();
		// List pkValues = rds.getFkValuesFromObject(obj, mif);
		Object[] pkVals = rds.getForeignKeyValues(obj, mif);
		Identity id = new Identity(referencedClass, pkVals);
		Class referencedProxy = rds.getItemProxyClass();
		if (referencedProxy != null)
		{
			try
			{
				return VirtualProxy.createProxy(referencedProxy, id);
			}
			catch (Exception ex)
			{
				ex.printStackTrace();
				return null;
			}
		}
		else
		{
			return getObjectByIdentity(id);
		}
	}

	/**
	 * Abort and close the transaction.
	 * Calling abort abandons all persistent object modifications and releases the
	 * associated locks.
	 * If transaction is not in progress a TransactionNotInProgressException is thrown
	 */
	public synchronized void abortTransaction() throws org.odmg.TransactionNotInProgressException
	{
		if (!inTransaction)
		{
			throw new org.odmg.TransactionNotInProgressException();
		}
		inTransaction = false;
		Enumeration allConnections = m_ConnectionManager.enumeration();
		while (allConnections.hasMoreElements())
		{
			try
			{
				Connection conn = (Connection) allConnections.nextElement();
				conn.rollback();
				conn.setAutoCommit(true);
			}
			catch (Throwable t)
			{
				System.out.println("Error in Transaction abort: " + t.getMessage());
			}
		}
	}

	/**
	 * begin a transaction against the underlying RDBMS.
	 * Calling <code>beginTransaction</code> multiple times,
	 * without an intervening call to <code>commitTransaction</code> or <code>abortTransaction</code>,
	 * causes the exception <code>TransactionInProgressException</code> to be thrown
	 * on the second and subsequent calls.
	 */
	public synchronized void beginTransaction() 
			throws org.odmg.TransactionInProgressException, org.odmg.TransactionAbortedException
	{
		if (inTransaction)
		{
			throw new org.odmg.TransactionInProgressException();
		}
		inTransaction = true;
		Enumeration allConnections = m_ConnectionManager.enumeration();
		while (allConnections.hasMoreElements())
		{
			try
			{
				Connection conn = (Connection) allConnections.nextElement();
				conn.setAutoCommit(false);
			}
			catch (Throwable t)
			{
				abortTransaction();
				throw new org.odmg.TransactionAbortedException("Transaction aborted: " 
						+ t.getMessage());
			}
		}
	}

	/**
	 * Commit and close the transaction.
	 * Calling <code>commit</code> commits to the database all
	 * UPDATE, INSERT and DELETE statements called within the transaction and
	 * releases any locks held by the transaction.
	 * If beginTransaction() has not been called before a
	 * TransactionNotInProgressException exception is thrown.
	 * If the transaction cannot be commited a TransactionAbortedException exception is thrown.
	 */
	public synchronized void commitTransaction() 
			throws org.odmg.TransactionNotInProgressException, org.odmg.TransactionAbortedException
	{
		if (!inTransaction)
		{
			throw new org.odmg.TransactionNotInProgressException();
		}
		inTransaction = false;
		Enumeration allConnections = m_ConnectionManager.enumeration();
		while (allConnections.hasMoreElements())
		{
			try
			{
				Connection conn = (Connection) allConnections.nextElement();
				conn.commit();
				conn.setAutoCommit(true);
			}
			catch (Throwable t)
			{
				abortTransaction();
				throw new org.odmg.TransactionAbortedException("Transaction aborted: " 
						+ t.getMessage());
			}
		}
	}

    public synchronized void checkPoint()
			throws org.odmg.TransactionNotInProgressException, org.odmg.TransactionAbortedException
	{
		Enumeration allConnections = m_ConnectionManager.enumeration();
		while (allConnections.hasMoreElements())
		{
			try
			{
				Connection conn = (Connection) allConnections.nextElement();
				conn.commit();
			}
			catch (Throwable t)
			{
			    System.out.println("checkpoint failed!");
			}
		}
	}


	/**
	 * retrieve a collection of itemClass Objects matching the Query query
	 * @param collectionClass type the collection to be returned
	 * @param itemClass Class of item in collection
	 * @param itemProxy Proxy-Class encapsulating itemClass instances, if null, no Proxies are used.
	 * @param vecFields the Vector of fields representing the foreign key relationship
	 * @param vecValues the Vector of values of the fields in vecFields.
	 */
	private ManageableCollection getCollectionByQuery(java.lang.Class collectionClass, 
			java.lang.Class itemClass, java.lang.Class itemProxy, 
			Query query) throws ClassNotPersistenceCapableException
	{
		if (_debug)
		{
			System.out.println("getCollectionByQuery (" + collectionClass + ", " + itemClass + ", " 
							   + query + ")");
		}
		ClassDescriptor mif = m_DescriptorRepository.getDescriptorFor(itemClass);
		ManageableCollection result = null;
		try
		{
			// we will add the queried elements to this collection:
			result = (ManageableCollection) collectionClass.newInstance();
			// now iterate over all elements and add them to the new collection
			Iterator i = getIteratorFromQuery(query, mif, itemProxy);
			while (i.hasNext())
			{
				result.ojbAdd(i.next());
			}
		}
		catch (InstantiationException ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		catch (IllegalAccessException ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		return result;
	}

	/**
	 * retrieve a collection of itemClass Objects matching the Query query
	 * @param itemClass Class of item in collection
	 * @param itemProxy Proxy-Class encapsulating itemClass instances, if null, no Proxies are used.
	 * @param vecFields the Vector of fields representing the foreign key relationship
	 * @param vecValues the Vector of values of the fields in vecFields.
	 */
	private java.util.Collection getCollectionByQuery(java.lang.Class itemClass, 
			java.lang.Class itemProxy, Query query) throws ClassNotPersistenceCapableException
	{
		if (_debug)
		{
			System.out.println("getCollectionByQuery " + itemClass + ", " + query);
		}
		ClassDescriptor mif = m_DescriptorRepository.getDescriptorFor(itemClass);
		Vector res = new Vector(100);
		Iterator i = getIteratorFromQuery(query, mif, itemProxy);
		while (i.hasNext())
		{
			res.add(i.next());
		}
		return res;
	}

	/**
	 * 
	 * retrieve a collection of itemClass Objects matching the Query query
	 * 
	 */
	public ManageableCollection getCollectionByQuery(Class collectionClass, 
			Query query) throws PersistenceBrokerException
	{
		Class itemClass = query.getSearchClass();
		ClassDescriptor cld = m_DescriptorRepository.getDescriptorFor(itemClass);
		ManageableCollection result = null;
		Class proxyClass = cld.getProxyClass();
		// if class is NOT an interface we can start searching its directly mapped table
		if (!cld.isInterface())
		{
			result = getCollectionByQuery(collectionClass, itemClass, proxyClass, query);
		}
		// if class is an extent, we have to search through all extent classes too
		if (cld.isExtent())
		{
			Vector extentClasses = cld.getExtentClasses();
			for (int i = 0; i < extentClasses.size(); i++)
			{
				Class ec = (Class) extentClasses.get(i);
				cld = m_DescriptorRepository.getDescriptorFor(ec);
				itemClass = cld.getClassOfObject();
				proxyClass = cld.getProxyClass();
				ManageableCollection tmp = getCollectionByQuery(collectionClass, itemClass, 
						proxyClass, query);
				if (result == null)
				{
					result = tmp;
				}
				else
				{
					result.ojbAddAll(tmp);
				}
			}
		}
		return result;
	}

	/**
	 * 
	 * retrieve a collection of itemClass Objects matching the Query query
	 * 
	 */
	public java.util.Collection getCollectionByQuery(Query query) throws PersistenceBrokerException
	{
		Class itemClass = query.getSearchClass();
		ClassDescriptor cld = m_DescriptorRepository.getDescriptorFor(itemClass);
		Collection result = null;
		Class proxyClass = cld.getProxyClass();
		// if class is NOT an interface we can start searching its directly mapped table
		if (!cld.isInterface())
		{
			result = getCollectionByQuery(itemClass, proxyClass, query);
		}
		// if class is an extent, we have to search through all extent classes too
		if (cld.isExtent())
		{
			Vector extentClasses = cld.getExtentClasses();
			for (int i = 0; i < extentClasses.size(); i++)
			{
				Class ec = (Class) extentClasses.get(i);
				cld = m_DescriptorRepository.getDescriptorFor(ec);
				itemClass = cld.getClassOfObject();
				proxyClass = cld.getProxyClass();
				Collection tmp = getCollectionByQuery(itemClass, proxyClass, query);
				if (result == null)
				{
					result = tmp;
				}
				else
				{
					result.addAll(tmp);
				}
			}
		}
		return result;
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param oid
	 * 
	 * @return
	 * 
	 * @throws ClassNotPersistenceCapableException
	 * 
	 * @see
	 */
	private Object getDBObject(Identity oid) throws ClassNotPersistenceCapableException
	{
		Class c = oid.getObjectsClass();
		ClassDescriptor mif = m_DescriptorRepository.getDescriptorFor(c);
		Object newObj = null;
		// Class is NOT an Interface: it has a directly mapped table and we lookup this table first:
		if (!mif.isInterface())
		{
			// 1. try to retrieve skalar fields from directly mapped table columns
			newObj = m_DbAccess.materializeObject(mif, oid);
		}
		// if we did not find the object yet AND if the mif represents an Extent,
		// we can lookup all tables of the extent classes:
		if (newObj == null && mif.isExtent())
		{
			Vector extentClasses = mif.getExtentClasses();
			for (int i = 0; i < extentClasses.size(); i++)
			{
				Class ec = (Class) extentClasses.get(i);
				mif = m_DescriptorRepository.getDescriptorFor(ec);
				// 1a. try to retrieve skalar fields from extent table columns
				newObj = m_DbAccess.materializeObject(mif, oid);
				if (newObj != null)
				{
					break;    // found object in one of the extent classes
				}
			}
		}
		// loading references is useful only when the Object could be found in db:
		if (newObj != null)
		{
			// cache object immediately , so that references
			// can be established from referenced Objects back to this Object
			m_ObjectCache.cache(oid, newObj);
			// 2. retrieve non-skalar fields that contain objects retrievable from other tables
			retrieveReferences(newObj, mif);
			// 3. retrieve collection fields from foreign-key related tables:
			retrieveCollections(newObj, mif);
		}
		return newObj;
	}



	/**
	 * 
	 * returns an Iterator that iterates Objects of class c if calling the .next()
	 * 
	 * method. The Elements returned come from a SELECT ... WHERE Statement
	 * 
	 * that is defined by the Query query.
	 * 
	 * If itemProxy is null, no proxies are used.
	 * 
	 */
	public java.util.Iterator getIteratorByQuery(Query query) throws PersistenceBrokerException
	{
		Class itemClass = query.getSearchClass();
		ClassDescriptor cld = m_DescriptorRepository.getDescriptorFor(itemClass);
		Class proxyClass = cld.getProxyClass();
		return getIteratorFromQuery(query, cld, proxyClass);
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param query
	 * @param mif
	 * @param itemProxyClass
	 * 
	 * @return
	 * 
	 * @see
	 */
	private Iterator getIteratorFromQuery(Query query, ClassDescriptor mif, Class itemProxyClass)
	{
		try
		{
			RsIterator iter = null;
			if (itemProxyClass == null)
			{
				// if no proxyClass is defined use an Iteror that marializes
				// the Objects represented by the resultset immediately.
				iter = new RsIterator(query, mif, this);
			}
			else
			{
				// if a proxyClass is defined use an Iterator that returns
				// proxy objects and thus defers materialization.
				iter = new ProxyRsIterator(query, mif, itemProxyClass, this);
			}
			return iter;
		}
		catch (Exception ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		return null;
	}

	/**
	 * 
	 * retrieve an instance of Class c from the underlying persistence system
	 * 
	 * by using a sample object for providing primary key information.
	 * 
	 */
	private java.lang.Object getObjectByExample(java.lang.Object exampleObj) 
			throws PersistenceBrokerException
	{
		Identity oid = new Identity(exampleObj);
		return getObjectByIdentity(oid);
	}

	/**
	 * Method declaration
	 * 
	 * 
	 * @param id
	 * 
	 * @return
	 * 
	 * @throws PersistenceBrokerException
	 * 
	 * @see
	 */
	public Object getObjectByIdentity(Identity id) throws PersistenceBrokerException
	{
		if (_debug)
		{
			System.out.println("getObjectByIdentity " + id);
			// check if object is present in ObjectCache:
		}
		Object obj = m_ObjectCache.lookup(id);
		// only perform a db lookup if necessary (object not cached yet)
		if (obj == null)
		{
			obj = getDBObject(id);
		}
		return obj;
	}

	/**
	 * retrieve an Object by query
	 * I.e perform a SELECT ... FROM ... WHERE ...  in an RDBMS
	 */
	public java.lang.Object getObjectByQuery(Query query) throws PersistenceBrokerException
	{
		if (query instanceof QueryByExample)
		{
			Object obj = query.getExampleObject();
			if (obj instanceof Identity)
			{
				return getObjectByIdentity((Identity) obj);
			}
			else
			{
				return getObjectByExample(obj);
			}
		}
		else
		{
			Vector coll = (Vector) getCollectionByQuery(query);

            try
            {
                return coll.get(0);
            }
            catch(Exception e)
            {
                return null;
            }
		}
	}

	/**
	 * 
	 * returns an Enumeration of PrimaryKey Objects for objects of class DataClass.
	 * 
	 * The Elements returned come from a SELECT ... WHERE Statement
	 * 
	 * that is defined by the fields and their coresponding values of vecFields
	 * 
	 * and vecValues.
	 * 
	 * Useful for EJB Finder Methods...
	 * 
	 * @param PrimaryKeyClass the pk class for the searched objects
	 * 
	 * @param query the query
	 * 
	 */
	public java.util.Enumeration getPKEnumerationByQuery(Class primaryKeyClass, 
			Query query) throws PersistenceBrokerException
	{
		if (_debug)
		{
			System.out.println("getPKEnumerationByQuery " + query);
		}
		ClassDescriptor cld = m_DescriptorRepository.getDescriptorFor(query.getSearchClass());
		return new PkEnumeration(query, cld, primaryKeyClass, this);
	}

	/**
	 * 
	 * makes object obj persistent in the underlying persistence system.
	 * 
	 * E.G. by INSERT INTO ... or UPDATE ...  in an RDBMS.
	 * 
	 * The ModificationState parameter can be used to generate optimized SQL code.
	 * 
	 * This functionality is typically called from transaction managers, that
	 * 
	 * track which objects have to be stored. Thus this store method does not
	 * 
	 * use update cascading to referenced objects.
	 * 
	 */
	public void store(Object obj, ObjectModification mod) throws PersistenceBrokerException
	{
		/*
		 * 
		 * if (_debug)
		 * 
		 * System.out.println("store " + obj + ", " + mod);
		 * 
		 * if (mod.needsInsert())
		 * 
		 * store(obj, true);
		 * 
		 * if (mod.needsUpdate())
		 * 
		 * store(obj, false);
		 * 
		 */
		store(obj);
	}

	/**
	 * 
	 * makes object obj persistent in the underlying persistence system.
	 * 
	 * E.G. by INSERT INTO ... or UPDATE ...  in an RDBMS.
	 * 
	 * The ModificationState parameter can be used to generate optimized SQL code.
	 * 
	 * This functionality is typically called from transaction managers, that
	 * 
	 * track which objects have to be stored. Thus this store method does not
	 * 
	 * use update cascading to referenced objects.
	 * 
	 */
	private void store(Object obj, boolean insert) throws PersistenceBrokerException
	{
		if (_debug)
		{
			System.out.println("store " + obj + ", " + insert);
			// proxies have to be stored only when their real subjects have been materialized
		}
    	if (obj != null)
    	{
    		if (obj instanceof VirtualProxy)
    		{
    			VirtualProxy proxy = (VirtualProxy) obj;
    			if (proxy.alreadyMaterialized())
    			{
    				obj = proxy.getRealSubject();
    			}
    			else
    			{
    				return;
    			}
    		}
    		ClassDescriptor mif = m_DescriptorRepository.getDescriptorFor(obj.getClass());
    		// 1. store primitive typed attributes
    		// if obj not present in db use INSERT
    		if (insert)
    		{
    			m_DbAccess.executeInsert(mif, obj);
    		}
    		// else use UPDATE
    		else
    		{
    			m_DbAccess.executeUpdate(mif, obj);
    		}
    		// cache object for symmetry with getObjectByXXX()
    		m_ObjectCache.cache(obj);
    		// 2. store references
    		storeReferences(obj, mif.getObjectReferenceDescriptors());
    		// 3. store dependend objects
    		storeCollections(obj, mif.getCollectionDescriptors());
    	}
    	else return;
	}

	/**
	 * this flag tells if the Broker is currently processing a Transaction
	 */
	private boolean inTransaction = false;

	/**
	 * returns true if the broker is currently running a transaction.
	 * @return boolean
	 */
	public boolean isInTransaction()
	{
		return inTransaction;
	}

	/**
	 * 
	 * Insert the method's description here.
	 * 
	 * Creation date: (07.02.2001 20:48:31)
	 * 
	 * @return ojb.broker.query.Query
	 * 
	 * @param cld ojb.broker.metadata.ClassDescriptor
	 * 
	 * @param fields java.util.Vector
	 * 
	 * @param values java.util.Vector
	 * 
	 */
	private Query getForeignKeyQuery(ClassDescriptor cld, FieldDescriptor[] fields, Object[] values)
	{
		Criteria criteria = new Criteria();
		for (int i = 0; i < fields.length; i++)
		{
			FieldDescriptor fld = (FieldDescriptor) fields[i];
			criteria.addEqualTo(fld.getPersistentField().getName(), values[i]);
		}
		criteria.orderBy(cld.getOrderby());
		Query result = QueryFactory.newQuery(cld.getClassOfObject(), criteria);
		return result;
	}

    public int getUniqueId(Class extent, String attribute)
    {
        return m_SequenceManager.getUniqueId(extent, attribute);
    }
}