/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.broker.accesslayer;
import java.sql.*;
import java.util.Enumeration;
import java.util.Vector;
import java.lang.reflect.Field;
import ojb.broker.metadata.*;
import ojb.broker.query.*;
import java.util.Iterator;
import java.lang.reflect.Constructor;
import ojb.broker.Identity;
import ojb.broker.ConversionStrategy;
import ojb.broker.PersistenceBrokerImpl;

/**
 * 
 * JdbcAccess is responsible for establishing performing
 * 
 * SQL Queries against remote Databases.
 * 
 * It hides all knowledge about JDBC from the BrokerImpl
 * 
 * 
 * 
 * @author <a href="mailto:thomas.mahler@itellium.com">Thomas Mahler<a>
 * 
 */
public class JdbcAccess
{

	/**
	 * en/disables tracing
	 */
	private static boolean _debug = false;

	/**
	 * set the _debug flag, to en/disable tracing
	 */
	public static void setDebug(boolean b)
	{
		_debug = b;
	}

	/**
	 * the broker in use
	 */
	private PersistenceBrokerImpl broker;

	/**
	 * constructor is private, use getInstance to get
	 * 
	 * the singleton instance of this class
	 * 
	 */
	public JdbcAccess(PersistenceBrokerImpl broker)
	{
	    this.broker = broker;
	}

   


	/**
	 * performs a DELETE operation against RDBMS.
	 * 
	 * @param oid The row identified by oid is deleted.
	 * 
	 * @param mif ClassDescriptor providing mapping information.
	 * 
	 */
	public boolean executeDelete(ClassDescriptor mif, Identity oid)
	{
		if (_debug)
		{
			System.out.println("executeDelete: " + oid);
		}
		try
		{
			PreparedStatement stmt = StatementManager.getDeleteStatement(mif, broker);
			synchronized (stmt)
			{
				StatementManager.bindDelete(stmt, oid);
				stmt.executeUpdate();
			}
			stmt = null;
			return true;
		}
		catch (SQLException ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
			return false;
		}
	}

	/**
	 * performs an INSERT operation against RDBMS.
	 * 
	 * @param obj The Object to be inserted as a row of the underlying table.
	 * 
	 * @param mif ClassDescriptor providing mapping information.
	 * 
	 */
	public boolean executeInsert(ClassDescriptor mif, Object obj)
	{
		if (_debug)
		{
			System.out.println("executeInsert: " + obj);
		}
		try
		{
			PreparedStatement stmt = StatementManager.getInsertStatement(mif,broker);
			synchronized (stmt)
			{
				StatementManager.bindInsert(stmt, mif, obj);
				stmt.executeUpdate();
			}
			stmt = null;
			return true;
		}
		catch (SQLException ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
			return false;
		}
	}

	/**
	 * performs a SELECT operation against RDBMS.
	 * @param qry the query string.
	 * @param mif ClassDescriptor providing JDBC information.
	 */
	ResultSet executeQuery(Query query, ClassDescriptor mif)
	{
		if (_debug)
		{
			System.out.println("executeQuery: " + query);
		}
		try
		{
			// if query has criteria, use them in where clause
			if (query.getCriteria() != null && ! query.getCriteria().isEmpty())
			{
				
				String sql = SqlGenerator.getInstance().getPreparedSelectStatement(query, mif);
				PreparedStatement stmt = StatementManager.getPreparedStatement(mif, sql, broker);
				StatementManager.bindStatement(stmt, query.getCriteria(), mif, 1);
				ResultSet rs = stmt.executeQuery();
				// as we return the resultset for further operations, we cannot release the statement yet.
				// that has to be done by the JdbcAccess-clients (i.e. RsIterator, ProxyRsIterator and PkEnumeration.)
				return rs;
				
			}
			// if query has no criteria, perform select * from table
			else
			{
				Statement stmt = StatementManager.getGenericStatement(mif, broker);
				String sql = SqlGenerator.getInstance().getSelectStatementDep(query, mif);
				ResultSet rs = stmt.executeQuery(sql);
				return rs;
			}
		}
		catch (Exception ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
			return null;
		}
	}

	/**
	 * performs an UPDATE operation against RDBMS.
	 * 
	 * @param obj The Object to be updated in the underlying table.
	 * 
	 * @param mif ClassDescriptor providing mapping information.
	 * 
	 */
	public boolean executeUpdate(ClassDescriptor mif, Object obj)
	{
		if (_debug)
		{
			System.out.println("executeUpdate: " + obj);
		}
		try
		{
			PreparedStatement stmt = StatementManager.getUpdateStatement(mif, broker);
			synchronized (stmt)
			{
				StatementManager.bindUpdate(stmt, mif, obj);
				stmt.executeUpdate();
			}
			stmt = null;
			return true;
		}
		catch (SQLException ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
			return false;
		}
	}

	/**
	 * performs a primary key lookup operation against RDBMS and materializes
	 * 
	 * an object from the resulting row. Only skalar attributes are filled from
	 * 
	 * the row, references are not resolved.
	 * 
	 * @param oid contains the primary key info.
	 * 
	 * @param mif ClassDescriptor providing mapping information.
	 * 
	 * @return the materialized object, null if no matching row was found or if
	 * 
	 * any error occured.
	 * 
	 */
	public Object materializeObject(ClassDescriptor mif, Identity oid)
	{
		Object obj = null;
		ResultSet rs;
		try
		{
			PreparedStatement stmt = StatementManager.getSelectByPKStatement(mif, broker);
			synchronized (stmt)
			{
				StatementManager.bindSelect(stmt, oid);
				rs = stmt.executeQuery();
				// data available read object, else return null
				if (rs.next())
				{
					obj = readObjectFromResultSet(rs, mif);
				}
				else
				{
					obj = null;
				}
				rs.close();
			}
		}
		catch (Exception ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		rs = null;
		return obj;
	}

	/**
	 * materialize a single object, described by mif,
	 * from the first row of the ResultSet rs.
	 * There are two possible strategies:
	 * 1. The persistent class defines a public constructor with arguments matching the persistent
	 * scalar attributes of the class. In this case we build an array args of arguments from rs
	 * and call Constructor.newInstance(args) to build an object.
	 * 2. The persistent class does not provide such a constructor, but only a public default
	 * constructor. In this case we create an empty instance with Class.newInstance().
	 * This empty instance is then filled by calling Field::set(obj,getObject(matchingColumn))
	 * for each attribute.
	 * The second strategy needs n calls to Field::set() which are much more expensive
	 * than the filling of the args array in the first strategy.
	 * client applications should therefore define adequate constructors to benefit from
	 * performance gain of the first strategy.
	 */
	private Object readObjectFromResultSet(ResultSet rs, ClassDescriptor mif)
	{
		Object val = null;
		FieldDescriptor fmd = null;
		Constructor multiArgsConstructor = mif.getConstructor();
		ConversionStrategy conversion = mif.getConversionStrategy();
		try
		{
			// if the class has an multiargument constructor, use it to construct the new object
			if (multiArgsConstructor != null)
			{
				// 1. build argument array
				FieldDescriptor[] fields = mif.getFieldDescriptions();
				Object[] args = new Object[fields.length];
				for (int i = 0; i < fields.length; i++)
				{
					fmd = fields[i];
					val = rs.getObject(fmd.getColNo());
					
					args[i] = conversion.sqlToJava(val, fmd);
				}
				System.out.println();
				// 2. call the constructor to build the object
				return multiArgsConstructor.newInstance(args);
			}
			// if no such constructor exists, use default constructor and reflection to fill object
			else
			{
				// 1. create an empty Object (persistent classes need a public default constructor)
				Class c = mif.getClassOfObject();
				Object obj = c.newInstance();
				// 2. fill all scalar attributes of the new object
				FieldDescriptor[] fields = mif.getFieldDescriptions();
				for (int i = 0; i < fields.length; i++)
				{
					fmd = fields[i];
					val = rs.getObject(fmd.getColNo());
					fmd.getPersistentField().set(obj, conversion.sqlToJava(val, fmd));
				}
				return obj;
			}
		}
		catch (Exception ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
			return null;
		}
	}

}