/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.broker.accesslayer;
import java.util.Hashtable;
import java.util.Vector;
import java.lang.ref.SoftReference;
import ojb.broker.metadata.ClassDescriptor;
import ojb.broker.metadata.DescriptorRepository;
import java.util.List;
import ojb.broker.Identity;
import ojb.broker.ClassNotPersistenceCapableException;

/**
 * The ObjectCache stores all Objects loaded by the PersistenceBroker from a DB.
 * When the PersistenceBroker tries to get an Object by its Primary key values
 * it first lookups the cache if the object has been already loaded and cached.
 * 
 * Using an ObjectCache has several advantages:
 * - it increases performance as it reduces DB lookups.
 * - it allows to perform circular lookups (as by crossreferenced objects)
 * that would result in non-terminating loops without such a cache.
 * - it maintains the uniqueness of objects as any Db row will be mapped to
 * exactly one object.
 * 
 * The cache uses soft-references which allows objects (softly) referenced by
 * the cache to be reclaimed by the Java Garbage Collector when they are not
 * longer referenced elsewhere.
 * 
 * @author <a href="mailto:thomas.mahler@itellium.com">Thomas Mahler<a>
 */
public class ObjectCache
{

	/**
	 * debugging flag, false by default = no debugging
	 */
	private static boolean _debug = false;

	/**
	 * set the debugging flag to true to enable debugging
	 */
	public static void setDebug(boolean b)
	{
		_debug = b;
	}

	/**
	 * singleton instance
	 */
	private static ObjectCache _instance;

	/**
	 * return the singleton instance
	 */
	public static ObjectCache getInstance()
	{
		if (_instance == null)
		{
			_instance = new ObjectCache();
		}
		return _instance;
	}

	/**
	 * Constructor declaration
	 * 
	 * 
	 * @see
	 */
	private ObjectCache() {}

	/**
	 * Make object obj persistent to Objectcache.
	 * compute objects identity and use it as key for the hashmap
	 */
	public void cache(Object obj) throws ClassNotPersistenceCapableException
	{
		Identity oid = new Identity(obj);
		this.cache(oid, obj);
	}

	/**
	 * 
	 * clear ObjectCache. I.e. remove all entries for classes and objects.
	 * 
	 */
	public void clear()
	{
		objectTable.clear();
	}

	private static String DIDNT_USE = "didn't use ";

	/**
	 * the hashtable holding all cached object
	 */
	private Hashtable objectTable = new Hashtable();

	/**
	 * en/disables cache
	 */
	private static boolean USE_CACHE = true;
	private static String USED = "used ";

	/**
	 * 
	 * makes object persistent to the Objectcache.
	 * 
	 * I'm using soft-references to allow gc reclaim unused objects
	 * 
	 * even if they are still cached.
	 * 
	 */
	public void cache(Identity oid, Object obj)
	{
		if ((obj != null) && USE_CACHE)
		{
			SoftReference ref = new SoftReference(obj);
			objectTable.put(oid.toString(), ref);
			if (_debug)
			{
				System.out.println("caching " + oid);
			}
		}
	}

	/**
	 * 
	 * Insert the method's description here.
	 * 
	 * Creation date: (20.11.2000 11:12:42)
	 * 
	 * @return boolean
	 * 
	 */
	public static boolean isCacheUsed()
	{
		return USE_CACHE;
	}

	/**
	 * 
	 * Lookup object with Identity oid in objectTable.
	 * 
	 * returns null if no matching id is found
	 * 
	 */
	public Object lookup(Identity oid)
	{
		boolean usedCache = true;
		Object obj = null;
		SoftReference ref = (SoftReference) objectTable.get(oid.toString());
		if (ref != null)
		{
			obj = ref.get();
			if (obj == null)
			{
				objectTable.remove(oid);    // Soft-referenced Object reclaimed by GC
				if (_debug)
				{
					System.out.println(oid + " reclaimed by GC");
				}
			}
		}
		else
		{
			usedCache = false;
		}
		if (_debug)
		{
			System.out.println((usedCache ? USED : DIDNT_USE) + "cache to retrieve " + oid);
		}
		return obj;
	}

	/**
	 * 
	 * removes an Object from the cache.
	 * 
	 */
	public void remove(Identity oid)
	{
		objectTable.remove(oid.toString());
	}

	/**
	 * 
	 * sets the flag USE_CACHE to the specified value.
	 * 
	 * if USE_CACHE is true caching is enabled, if false caching is disabled
	 * 
	 * @param newUSE_CACHE boolean
	 * 
	 */
	public static void setCacheUsed(boolean newUSE_CACHE)
	{
		USE_CACHE = newUSE_CACHE;
	}

}