/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.broker.accesslayer;
import java.util.Iterator;
import java.sql.ResultSet;
import java.util.NoSuchElementException;
import java.util.Vector;
import java.util.Enumeration;
import java.lang.reflect.Field;
import ojb.broker.metadata.*;
import ojb.broker.*;
import java.sql.Connection;
import ojb.broker.query.*;
import java.sql.SQLException;
import java.lang.reflect.Constructor;

/**
 * RsIterator can be used to iterate over a jdbc ResultSet to retrieve
 * persistent objects step-by-step and not all at once.
 * In fact the PersistenceBroker::getCollectionByQuery(...) method
 * uses a RsIterator internally to build up a Collection of objects
 * 
 * NOTE: this code uses features that only JDBC 2.0 compliant Drivers
 * support. It will NOT work with JDBC 1.0 Drivers (e.g. SUN's JdbcOdbcDriver)
 * If you are forced to use such a driver, you can use code from
 * the 0.1.30 release.
 * 
 * @author <a href="mailto:thomas.mahler@itellium.com">Thomas Mahler<a>
 */
public class RsIterator implements Iterator
{

	/**
	 * debugging flag, false by default = no debugging
	 */
	protected static boolean _debug = false;

	protected static ObjectCache cache = ObjectCache.getInstance();

	/**
	 * set the debugging flag to true to enable debugging
	 */
	public static void setDebug(boolean b)
	{
		_debug = b;
	}

	/**
	 * reference to the PersistenceBroker
	 */
	protected PersistenceBrokerImpl m_broker;

	/**
	 * the underlying resultset
	 */
	protected ResultSet m_rs;

	/**
	 * the classdescriptor providing the mapping information
	 */
	protected ClassDescriptor m_mif;

/**
 * returns true if there are still more rows in the underlying ResultSet.
 * Returns false if ResultSet is exhausted.
 */
public synchronized boolean hasNext()
{
	try
	{
		if (!hasCalledCheck)
		{
			hasCalledCheck = true;
			hasNext = m_rs.next();
			if (!hasNext)
			{
				m_rs.close();
				m_rs = null;
			}

		}
	}
	catch (Exception ex)
	{
		hasNext = false;
	}
	finally
	{
		if (_debug)
			System.out.println("hasNext() -> " + hasNext);
		return hasNext;
	}
}

	/**
	 * moves to the next row of the underlying ResultSet and
	 * returns the corresponding Object materialized from this row.
	 */
public synchronized Object next() throws NoSuchElementException {
	try
	{
	    if (! hasCalledCheck) hasNext();
	    hasCalledCheck = false;
	    if (hasNext)
	    {
		    Object obj = getObjectFromResultSet();
		    if (_debug)
			    System.out.println("next() -> " + obj);
		    return obj;
        }
        else throw new NoSuchElementException();
	}
	catch (Exception ex)
	{
		if (_debug)
			System.out.println("next() -> ERROR");
		ex.printStackTrace();
		throw new NoSuchElementException();
	}
}     

	/**
	 * removing is not supported
	 */
	public void remove()
	{
		System.out.println("removing not supported by RsIterator");
	}

	/**
	 * 
	 * RsIterator constructor.
	 * 
	 * @param qry the SELECT producing the underlying resultset
	 * 
	 * @param mif ojb.broker.metadata.ClassDescriptor
	 * 
	 */
	public RsIterator(Query query, ClassDescriptor mif, PersistenceBrokerImpl broker)
	{
		if (_debug)
		{
			System.out.println("RsIterator(" + query + ", " + mif + ")");
		}
		m_rs = new JdbcAccess(broker).executeQuery(query, mif);
		m_broker = broker;
		m_mif = mif;
	}

	/**
	 * returns an Identity object representing the current resultset row
	 */
	protected Identity getIdentityFromResultSet() throws PersistenceBrokerException
	{
		try
		{
			// 1. get an empty instance of the target class
			Class c = m_mif.getClassOfObject();
			Object obj = c.newInstance();
			// 2. fill primary key values from Resultset
			Object colValue;
			FieldDescriptor fld;
			FieldDescriptor[] pkfields = m_mif.getPkFields();
			for (int i = 0; i < pkfields.length; i++)
			{
				fld = pkfields[i];
				colValue = m_rs.getObject(fld.getColumnName());
				fld.getPersistentField().set(obj, colValue);
			}
			// 3. return a representing Identity
			return new Identity(c, m_mif.getKeyValues(obj));
		}
		catch (java.sql.SQLException ex)
		{
			throw new PersistenceBrokerDBException(ex.getMessage());
		}
		catch (Exception ex)
		{
			throw new ClassNotPersistenceCapableException("No public Constructor for " 
					+ m_mif.getClassOfObject() + "\n" + ex.getMessage());
		}
	}

	/**
	 * materialize a single object, described by mif,
	 * from the first row of the ResultSet rs.
	 * There are two possible strategies:
	 * 1. The persistent class defines a public constructor with arguments matching the persistent
	 * scalar attributes of the class. In this case we build an array args of arguments from rs
	 * and call Constructor.newInstance(args) to build an object.
	 * 2. The persistent class does not provide such a constructor, but only a public default
	 * constructor. In this case we create an empty instance with Class.newInstance().
	 * This empty instance is then filled by calling Field::set(obj,getObject(matchingColumn))
	 * for each attribute.
	 * The second strategy needs n calls to Field::set() which are much more expensive
	 * than the filling of the args array in the first strategy.
	 * client applications should therefore define adequate constructors to benefit from
	 * performance gain of the first strategy.
	 */
	private Object readObjectFromCurrentRow()
	{
		Object val = null;
		FieldDescriptor fmd = null;
		Constructor multiArgsConstructor = m_mif.getConstructor();
		ConversionStrategy conversion = m_mif.getConversionStrategy();
		try
		{
			// if the class has an multiargument constructor, use it to construct the new object
			if (multiArgsConstructor != null)
			{
				// 1. build argument array
				FieldDescriptor[] fields = m_mif.getFieldDescriptions();
				Object[] args = new Object[fields.length];
				for (int i = 0; i < fields.length; i++)
				{
					fmd = fields[i];
					val = m_rs.getObject(fmd.getColNo());

					args[i] = conversion.sqlToJava(val, fmd);
				}
				// 2. call the constructor to build the object
				return multiArgsConstructor.newInstance(args);
			}
			// if no such constructor exists, use default constructor and reflection to fill object
			else
			{
				// 1. create an empty Object (persistent classes need a public default constructor)
				Class c = m_mif.getClassOfObject();
				Object obj = c.newInstance();
				// 2. fill all scalar attributes of the new object
				FieldDescriptor[] fields = m_mif.getFieldDescriptions();
				for (int i = 0; i < fields.length; i++)
				{
					fmd = fields[i];
					val = m_rs.getObject(fmd.getColNo());
					fmd.getPersistentField().set(obj, conversion.sqlToJava(val, fmd));
				}
				return obj;
			}
		}
		catch (Exception ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
			return null;
		}
	}


	/**
	 * returns a fully materialized Object from the current row of the
	 * underlying resultset.
	 * Works as follows:
	 * - read primary key values from current row
	 * - build and Identity from the primary key values
	 * - ask the PersistenceBroker to retrieve the Object identified by this Identity
	 */
	private Object getObjectFromResultSet()
	{
		try
		{
		    // 1.materialize Object with primitive attributes filled from current row
		    Object result = this.readObjectFromCurrentRow();

		    // 2. check if Object is in cache. if so return cached version.
		    //    If Object is not in cache fill reference and collection attributes
		    Identity oid = new Identity(result);
		    Object check = cache.lookup(oid);
		    if (check != null) return check;
		    else
		    {
		        cache.cache(oid, result);
		        m_broker.retrieveReferences(result, m_mif);
		        m_broker.retrieveCollections(result, m_mif);
		        return result;
            }
		}
		catch (Exception ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		return null;
	}

	/**
	 * Release all internally used Database resources of the iterator.
	 * Clients must call this methods explicitely if the iterator is not
	 * exhausted by the client application. If the Iterator is exhauseted
	 * this method will be called implicitely.
	 */
	public void releaseDbResources()
	{
		try
		{
		    m_rs.getStatement().close();
			m_rs.close();
			m_rs = null;
		}
		catch (SQLException ex) {}
	}

	protected boolean hasCalledCheck = false;
	protected boolean hasNext = false;
}