/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.broker.accesslayer;
import java.sql.PreparedStatement;
import java.sql.Statement;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import ojb.broker.metadata.ClassDescriptor;
import ojb.broker.metadata.DescriptorRepository;
import ojb.broker.metadata.JdbcConnectionDescriptor;
import ojb.broker.PersistenceBrokerImpl;

/**
 * holds Statements that are used for persistence operations on a given class.
 * @author: Thomas Mahler
 */
class StatementsForClass
{

	/**
	 * the managed class
	 */
	private Class clazz;

	/**
	 * the classes mapping descriptor
	 */
	private ClassDescriptor cld;

	/**
	 * the JDBC Connection used for this class
	 */
	private Connection connection;

	/**
	 * the statement used for PK lookups
	 */
	private PreparedStatement selectByPKkStmt;

	/**
	 * the statement used for inserts
	 */
	private PreparedStatement insertStmt;

	/**
	 * the statement used for updates
	 */
	private PreparedStatement updateStmt;

	/**
	 * the statement used for deletes
	 */
	private PreparedStatement deleteStmt;

	/**
	 * the generator used sql generation
	 */
	private SqlGenerator sqlGenerator;

	/** sets the scrollmode of JDBC ResultSets. OJB needs Forward_only only, but
	  * for certain RDBMS and JDBC drivers other modes may be necessary.
	  */
	public static int TYPE = ResultSet.TYPE_SCROLL_INSENSITIVE;

	/** sets the updatabe feature of JDBC ResultSets. OJB does not
	  * need updatable ResultSets, but for certain RDBMS or JDBC drivers
	  * it may be necessary to change this value.
	  */
	public static int CONCUR = ResultSet.CONCUR_READ_ONLY;

	/** sets the escape processing mode*/
	public static boolean ESCAPEPROCESSING = false;

	/**
	 *
	 * StatementsForClass constructor comment.
	 *
	 */
	private StatementsForClass()
	{
		super();
	}

	/**
	 * StatementsForClass constructor.
	 */
	public StatementsForClass(ClassDescriptor mif, PersistenceBrokerImpl broker)
	{
		cld = mif;
		clazz = mif.getClassOfObject();
		ConnectionManager manager = new ConnectionManager(broker);
		connection = manager.getConnectionForClassDescriptor(mif);
		sqlGenerator = SqlGenerator.getInstance();
	}

    /**
	 * prepares a statement with parameters that should work with most RDBMS
	 */
    private PreparedStatement prepareStatement(String sql) throws java.sql.SQLException
    {

        PreparedStatement result = null;
        // if a JDBC1.0 driver is used the signature
        // prepareStatement(String, int, int) is  not defined.
        // we then call the JDBC1.0 variant prepareStatement(String)
        try
        {
            result = connection.prepareStatement(connection.nativeSQL(sql), TYPE, CONCUR);
        }
        catch(AbstractMethodError err)
        {
            result = connection.prepareStatement(connection.nativeSQL(sql));
        }
        result.setEscapeProcessing(ESCAPEPROCESSING);
        return result;
    }

    /**
	 * creates a statement with parameters that should work with most RDBMS
	 */
    private Statement createStatement() throws java.sql.SQLException
    {
        Statement result = null;
        // if a JDBC1.0 driver is used the signature
        // createStatement(int, int) is  not defined.
        // we then call the JDBC1.0 variant createStatement()
        try
        {
            result = connection.createStatement(TYPE, CONCUR);
        }
        catch(AbstractMethodError err)
        {
            result = connection.createStatement();
        }
        result.setEscapeProcessing(ESCAPEPROCESSING);
        return result;
    }

	/**
	 * returns the DELETE Statement used for clazz.
	 * @return java.sql.PreparedStatement
	 */
	public PreparedStatement getDeleteStmt()
	{
		if (deleteStmt == null)
		{
			synchronized (this)
			{
				String sql = sqlGenerator.getPreparedDeleteStatement(cld);
				// System.out.println("prepare: " + sql);
				try
				{
					deleteStmt = prepareStatement(sql);
				}
				catch (java.sql.SQLException ex)
				{
					ex.printStackTrace();
				}
			}
		}
		return deleteStmt;
	}

	/**
	 * returns a generic unprepared Statement used for clazz.
	 * @return java.sql.Statement
	 */
	public Statement getGenericStmt()
	{
		Statement stmt = null;
		try
		{
			stmt = createStatement();
		}
		catch (java.sql.SQLException ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		return stmt;
	}

	/**
	 * returns the INSERT Statement used for clazz.
	 * @return java.sql.PreparedStatement
	 */
	public PreparedStatement getInsertStmt()
	{
		if (insertStmt == null)
		{
			synchronized (this)
			{
				String sql = sqlGenerator.getPreparedInsertStatement(cld);
				// System.out.println("prepare: " + sql);
				try
				{
					insertStmt = prepareStatement(sql);
				}
				catch (java.sql.SQLException ex)
				{
					ex.printStackTrace();
				}
			}
		}
		return insertStmt;
	}

	/**
	 * returns a prepared Statement used for clazz.
	 * @return java.sql.Statement
	 */
	public PreparedStatement getPreparedStmt(String sql)
	{
		PreparedStatement stmt = null;
		try
		{
			stmt = prepareStatement(sql);
		}
		catch (java.sql.SQLException ex)
		{
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		return stmt;
	}

	/**
	 * returns the SELECT Statement used for clazz.
	 * @return java.sql.PreparedStatement
	 */
	public PreparedStatement getSelectByPKkStmt()
	{
		if (selectByPKkStmt == null)
		{
			synchronized (this)
			{
				String sql = sqlGenerator.getPreparedSelectByPkStatement(cld);
				// System.out.println("prepare: " + sql);
				try
				{
					selectByPKkStmt = prepareStatement(sql);
				}
				catch (java.sql.SQLException ex)
				{
					ex.printStackTrace();
				}
			}
		}
		return selectByPKkStmt;
	}

	/**
	 * returns the UPDATE Statement used for clazz.
	 * @return java.sql.PreparedStatement
	 */
	public PreparedStatement getUpdateStmt()
	{
		if (updateStmt == null)
		{
			synchronized (this)
			{
				String sql = sqlGenerator.getPreparedUpdateStatement(cld);
				// System.out.println("prepare: " + sql);
				try
				{
					updateStmt = prepareStatement(sql);
				}
				catch (java.sql.SQLException ex)
				{
					ex.printStackTrace();
				}
			}
		}
		return updateStmt;
	}

}