package ojb.broker.server;

import ojb.broker.*;
import ojb.broker.query.Query;
import org.odmg.TransactionNotInProgressException;
import org.odmg.TransactionInProgressException;
import org.odmg.TransactionAbortedException;
import java.util.Collection;
import java.util.Iterator;
import java.util.Enumeration;

import java.net.*;
import java.io.*;
import java.util.zip.*;

public class PersistenceBrokerClient implements PersistenceBroker, Serializable
{

    private static final int invalidId = -1;

    private ServerPool servers;

    private ServerEntry currentServer;

    private int brokerId = invalidId;

    public PersistenceBrokerClient(ServerPool serverpool)
    {
        servers = serverpool;
        currentServer = servers.roundRobin();
    }

    private Object callServer(int methodIndex)
    {
       return callServer(methodIndex, brokerId, null, null);
    }

    private Object callServer(int methodIndex, Object arg1)
    {
       return callServer(methodIndex, brokerId, arg1, null);
    }

    private Object callServer(int methodIndex, Object arg1, Object arg2)
    {
       return callServer(methodIndex, brokerId, arg1, arg2);
    }

    private Object callServer(int methodIndex, int brokerid, Object arg1, Object arg2)
    {
        //System.out.println("callServer(" + methodIndex + ", " + brokerId + ", " + arg1 + ", " + arg2 + ")");
        Socket connection = null;
        try
        {
            connection = currentServer.getConnection();
            OutputStream outstream = connection.getOutputStream();
            BufferedOutputStream bos = new BufferedOutputStream(outstream);
            ObjectOutputStream oos = new ObjectOutputStream(bos);

            // this is our simple protocol:
            Request request = new Request(methodIndex, brokerid, arg1, arg2);
            oos.writeObject(request);
            //oos.writeInt(methodIndex);  // index of the method to call
            //oos.writeInt(brokerid);     // the broker in use by this client. -1 indicates: no broker selected yet
            //oos.writeObject(arg1);      // the methods first argument (may be null)
            //oos.writeObject(arg2);      // the methodssecond argument (may be null)

            oos.flush();
            outstream.flush();
            //connection.shutdownOutput();

            // now see what the server has done for us:
            InputStream instream = connection.getInputStream();
            BufferedInputStream bis = new BufferedInputStream(instream);
            ObjectInputStream ois = new ObjectInputStream(bis);
            Object result =  ois.readObject();

            //connection.shutdownInput();

            connection.close();

            return result;
        }
        catch (Throwable t)
        {
            try
            {
                System.out.println(t.getMessage());
                t.printStackTrace();
                connection.close();
            }
            catch(IOException e)
            {
                System.out.println(e.getMessage());
            }
            System.out.println(t.getMessage());
            return null;
        }
    }



    public void delete(Object obj) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.delete, obj);
        if (result instanceof PersistenceBrokerException)
        {
            throw (PersistenceBrokerException) result;
        }
        else if (result != null) throw new PersistenceBrokerException(((Throwable) result).getMessage());
    }

    public void store(Object obj) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.store, obj);
        if (result instanceof PersistenceBrokerException)
        {
            throw (PersistenceBrokerException) result;
        }
        else if (result != null) throw new PersistenceBrokerException(((Throwable) result).getMessage());
    }

    public void abortTransaction() throws TransactionNotInProgressException
    {
        Object result = callServer(PBMethodIndex.abortTransaction);
        // if -l: tx is aborted and the broker can be released!
        if (result instanceof Integer && ((Integer)result).intValue() == -1)
        {
            brokerId = -1;
        }
        else throw new TransactionNotInProgressException();
    }

    public void beginTransaction()
      throws TransactionInProgressException, TransactionAbortedException
    {
        // allow roundrobin request for each new transaction
        currentServer = servers.roundRobin();

        Object result = callServer(PBMethodIndex.beginTransaction);
        if (result instanceof Integer)
        {
            brokerId = ((Integer) result).intValue();
        }
        else if (result instanceof TransactionInProgressException )
        {
            throw (TransactionInProgressException) result;
        }
        else if (result instanceof TransactionAbortedException )
        {
            throw (TransactionAbortedException) result;
        }

    }

    public void commitTransaction()
      throws TransactionNotInProgressException, TransactionAbortedException
    {
        Object result = callServer(PBMethodIndex.commitTransaction);
        if (result instanceof Integer && ((Integer)result).intValue() == -1)
        {
            brokerId = -1;
        }
        else if (result instanceof TransactionAbortedException)
        {
            brokerId = -1;
            throw (TransactionAbortedException) result;
        }
        else
        {
            brokerId = -1;
            throw new TransactionNotInProgressException();
        }
    }

    public Collection getCollectionByQuery(Query query) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.getCollectionByQuery, query);
        if (result instanceof Collection) return (Collection) result;
        else if (result instanceof PersistenceBrokerException) throw (PersistenceBrokerException) result;
        else throw new PersistenceBrokerException(((Throwable) result).getMessage());
    }

    public ManageableCollection getCollectionByQuery(Class collectionClass, Query query)
      throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.getCollectionByQueryMC, collectionClass, query);
        if (result instanceof ManageableCollection) return (ManageableCollection) result;
        else if (result instanceof PersistenceBrokerException) throw (PersistenceBrokerException) result;
        else throw new PersistenceBrokerException(((Throwable) result).getMessage());
    }

    public Iterator getIteratorByQuery(Query query) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.getIteratorByQuery, query);
        if (result instanceof Iterator) return (Iterator) result;
        else if (result instanceof PersistenceBrokerException) throw (PersistenceBrokerException) result;
        else throw new PersistenceBrokerException(((Throwable) result).getMessage());
    }

    public Object getObjectByIdentity(Identity id) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.getObjectByIdentity, id);
        if (result instanceof PersistenceBrokerException) throw (PersistenceBrokerException) result;
        else if (result instanceof Throwable)
            throw new PersistenceBrokerException(((Throwable) result).getMessage());
        else return result;
    }

    public Object getObjectByQuery(Query query) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.getObjectByQuery, query);
        if (result instanceof PersistenceBrokerException) throw (PersistenceBrokerException) result;
        else if (result instanceof Throwable)
            throw new PersistenceBrokerException(((Throwable) result).getMessage());
        else return result;
    }

    public Enumeration getPKEnumerationByQuery(Class primaryKeyClass,
      Query query) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.getPkEnumerationByQuery, primaryKeyClass, query);
        if (result instanceof Enumeration) return (Enumeration) result;
        else if (result instanceof PersistenceBrokerException) throw (PersistenceBrokerException) result;
        else throw new PersistenceBrokerException(((Throwable) result).getMessage());

    }

    public void store(Object obj,
      ObjectModification modification) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.storeWithModification, obj, modification);
        if (result instanceof PersistenceBrokerException)
        {
            throw (PersistenceBrokerException) result;
        }
        else if (result != null) throw new PersistenceBrokerException(((Throwable) result).getMessage());

    }


    public void ping()
    {
        Object result = callServer(PBMethodIndex.ping);
        System.out.println(result.toString());
    }

    public static void main(String[] args)
    {
        String[] servers = {"localhost:2001"};
        ServerPool pool = new ServerPool(servers);
        PersistenceBrokerClient pb = new PersistenceBrokerClient(pool);

        for (int i=0; i<1000; i++)
        {
            pb.beginTransaction();
            pb.ping();

            pb.commitTransaction();
        }


    }

    public int getUniqueId(Class extent, String attribute) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.getUniqueId, extent, attribute);
        if (result instanceof Integer) return ((Integer) result).intValue();
        else if (result instanceof PersistenceBrokerException) throw (PersistenceBrokerException) result;
        else throw new PersistenceBrokerException(((Throwable) result).getMessage());
    }

    public boolean hasNext(int iteratorId) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.hasNext, new Integer(iteratorId));
        if (result instanceof Boolean) return ((Boolean) result).booleanValue();
        else throw new PersistenceBrokerException();
    }

    public Object next(int iteratorId) throws PersistenceBrokerException
    {
        Object result = callServer(PBMethodIndex.next, new Integer(iteratorId));
        if (result instanceof Exception) throw (PersistenceBrokerException) result;
        else return result;
    }

    public void releaseDbResources(int iteratorId)
    {
        Object result = callServer(PBMethodIndex.releaseDbResources, new Integer(iteratorId));
    }

}