package ojb.broker.server;

import java.net.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import org.odmg.*;
import org.xml.sax.SAXException;
import javax.xml.parsers.ParserConfigurationException;
import ojb.broker.PersistenceBroker;
import ojb.broker.PersistenceBrokerImpl;
import ojb.broker.PersistenceBrokerException;
import ojb.broker.query.Query;
import ojb.broker.Identity;
import ojb.broker.ObjectModification;
import ojb.broker.PersistenceBrokerFactory;
import ojb.broker.accesslayer.RsIterator;
import ojb.broker.accesslayer.PkIterator;

    
public class RequestProcessor implements Runnable {

   /** the repository file holding the O/R mapping information */
  private File repositoryFile;

  private PersistenceBrokerServer server;
  private BrokerPool brokerpool;
  private static RsIteratorTable iterators = new RsIteratorTable();
  private static final boolean DEBUG = false;

  private Socket connection;

  public synchronized void setSocket(Socket s)
  {
    connection = s;
    notify();
  }



  
  public RequestProcessor(PersistenceBrokerServer srv, BrokerPool pool)
  {
     connection = null;
     brokerpool = pool;
     server = srv;
  }
  

  public synchronized void run()
  {
    while (true) {
        if (connection == null)
        {
            if (DEBUG) System.out.println(Thread.currentThread().getName() + " waits...");
            try
            {
                wait();
            }
            catch (InterruptedException e)
            {
                continue;
            }
        }
        try
        {
            if (DEBUG) System.out.println(Thread.currentThread().getName() + " working...");
            dispatch(connection);
        }
        catch (Exception e)
        {
            System.out.println(e.getMessage());
            e.printStackTrace();
        }
        finally
        {
            try
            {
                connection.close();
                connection = null;
                server.releaseProcessor(this);

            }
            catch (IOException e)
            {
                connection = null;
                server.releaseProcessor(this);
            }
        }
    } // end while
  } // end run



    private void dispatch(Socket connection) throws IOException, ClassNotFoundException
    {
        connection.setSoTimeout(10000);
        connection.setTcpNoDelay(true);

        InputStream instream = connection.getInputStream();
        BufferedInputStream bis = new BufferedInputStream(instream);
        ObjectInputStream ois = new ObjectInputStream(bis);

        // read in the request sent from the client
        Request request = (Request) ois.readObject();
        int methodIndex = request.getMethodIndex();
        int brokerId = request.getBrokerId();
        Object  arg1 = request.getArg1();
        Object  arg2 = request.getArg2();

        //connection.shutdownInput();

        if (DEBUG)
        {
            System.out.print("dispatching [" + methodIndex + ",");
            System.out.print(brokerId + ",");
            System.out.print(arg1 + ",");
            System.out.println(arg2 + "]");
        }

        PersistenceBroker broker;
        Object result = null;

        switch (methodIndex)
        {
            case PBMethodIndex.beginTransaction:
            {
                // the client has no dedicated broker yet.
                // thus we select a currently unused broker to be used by this client
                // the client we send this broker id in subsequent calls.
                // This allows to keep the "state" of a session in the clients
                // dedicated broker
                if (brokerId == -1)
                {
                    brokerId = brokerpool.getFreeBrokerId();
                    result = new Integer(brokerId);
                    broker = brokerpool.getBrokerById(brokerId);
                    try
                    {
                        broker.beginTransaction();
                        System.out.println("Broker" + brokerId + " started tx" );
                    }
                    catch(TransactionInProgressException e)
                    {
                        System.out.println("Server ERR: told client to take an already used broker");
                        brokerpool.releaseBrokerId(brokerId);
                        result = new TransactionInProgressException();
                    }
                }
                else
                {   System.out.println("Client ERR: Client tries to open tx twice");
                    brokerpool.releaseBrokerId(brokerId);
                    result = new TransactionInProgressException();
                }
                break;
            }

            case PBMethodIndex.abortTransaction:
            {
                if (brokerId != -1)
                {
                    broker = brokerpool.getBrokerById(brokerId);
                    try
                    {
                        broker.abortTransaction();
                        brokerpool.releaseBrokerId(brokerId);
                        result = new Integer(-1);
                    }
                    catch(TransactionNotInProgressException e)
                    {
                        result = new TransactionNotInProgressException();
                    }
                }
                else result = new TransactionNotInProgressException();
                break;
            }

            case PBMethodIndex.commitTransaction:
            {
                if (brokerId != -1)
                {
                    broker = brokerpool.getBrokerById(brokerId);

                    try
                    {
                        broker.commitTransaction();
                        brokerpool.releaseBrokerId(brokerId);
                        result = new Integer(-1);
                    }
                    catch(TransactionNotInProgressException e)
                    {
                        brokerpool.releaseBrokerId(brokerId);
                        result = e;
                    }
                    catch(TransactionAbortedException e)
                    {
                        brokerpool.releaseBrokerId(brokerId);
                        result = e;
                    }
                }
                else result = new TransactionNotInProgressException();
                break;
            }

            case PBMethodIndex.delete:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);
                try
                {
                    broker.delete(arg1);
                    brokerpool.releaseBrokerId(brokerId);
                    result = null;
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.getCollectionByQueryMC:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);
                try
                {
                    result = broker.getCollectionByQuery((Class) arg1, (Query) arg2);
                    brokerpool.releaseBrokerId(brokerId);
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.getCollectionByQuery:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);
                try
                {
                    result = broker.getCollectionByQuery((Query) arg1);
                    brokerpool.releaseBrokerId(brokerId);
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.getIteratorByQuery:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);

                try
                {
                    // RsIterators contain open resultsets, we thus can serialize them,
                    // but have to generate stubs that tunnel their functionality
                    RsIterator rsiter = (RsIterator) broker.getIteratorByQuery((Query) arg1);
                    Integer iid = iterators.getUniqueId();
                    iterators.register(iid, rsiter);
                    result = new RsIteratorStub((PersistenceBrokerClient)PersistenceBrokerFactory.createPersistenceBroker(), iid.intValue());
                    brokerpool.releaseBrokerId(brokerId);
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.getObjectByIdentity:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);

                try
                {
                    result = broker.getObjectByIdentity((Identity) arg1);
                    brokerpool.releaseBrokerId(brokerId);
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                if (DEBUG) System.out.println("ObjectByIdentity: " + result);
                break;
            }

            case PBMethodIndex.getObjectByQuery:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);
                try
                {
                    result = broker.getObjectByQuery((Query) arg1);
                    brokerpool.releaseBrokerId(brokerId);
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.getPkEnumerationByQuery:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);
                try
                {
                    Enumeration enum = broker.getPKEnumerationByQuery((Class) arg1, (Query) arg2);
                    Iterator iter = new PkIterator(enum);
                    // RsIterators contain open resultsets, we thus can serialize them,
                    // but have to generate stubs that tunnel their functionality
                    Integer iid = iterators.getUniqueId();
                    iterators.register(iid, iter);
                    result = new PkEnumerationStub((PersistenceBrokerClient)PersistenceBrokerFactory.createPersistenceBroker(),iid.intValue());


                    brokerpool.releaseBrokerId(brokerId);
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.store:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);
                try
                {
                    broker.store(arg1);
                    ((PersistenceBrokerImpl)broker).checkPoint();
                    brokerpool.releaseBrokerId(brokerId);
                    result = null;
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.storeWithModification:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);

                try
                {
                    broker.store(arg1, (ObjectModification) arg2);
                    ((PersistenceBrokerImpl)broker).checkPoint();
                    brokerpool.releaseBrokerId(brokerId);
                    result = null;
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.getUniqueId:
            {
                if (brokerId == -1) brokerId = brokerpool.getFreeBrokerId();
                broker = brokerpool.getBrokerById(brokerId);

                try
                {
                    int uid = broker.getUniqueId((Class) arg1, (String) arg2);
                    if (DEBUG) System.out.println("UID: " + uid);
                    brokerpool.releaseBrokerId(brokerId);
                    result = new Integer(uid);
                }
                catch(PersistenceBrokerException e)
                {
                    brokerpool.releaseBrokerId(brokerId);
                    result = e;
                }
                break;
            }

            case PBMethodIndex.hasNext:
            {
                Integer id = (Integer) arg1;
                Iterator iter = iterators.getIteratorById(id);

                boolean hasnext = iter.hasNext();
                result = new Boolean(hasnext);
                if (!hasnext)
                {
                    iterators.release(id);
                }
                break;
            }

            case PBMethodIndex.next:
            {
                Integer id = (Integer) arg1;
                Iterator iter = iterators.getIteratorById(id);

                try
                {
                    result = iter.next();
                }
                catch(NoSuchElementException e)
                {
                    result = e;
                    iterators.release(id);
                }
                break;

            }

            case PBMethodIndex.releaseDbResources:
            {
                Integer id = (Integer) arg1;
                iterators.release(id);
                break;
            }

            case PBMethodIndex.ping:
            {
                result = "ping "  + connection.getLocalAddress() + ":" + connection.getPort() + " from " + connection.getInetAddress() + " " + Thread.currentThread().getName() + ", using broker: " + brokerId ;
                break;
            }


            default:
            {
                break;
            }

        }

        //if (DEBUG) System.out.println("sending result: " + result + " (" + ((result==null) ? "" : result.getClass().getName()) +")");

        // send result back to client
        OutputStream outstream = connection.getOutputStream();
        BufferedOutputStream bos = new BufferedOutputStream(outstream);
        ObjectOutputStream oos = new ObjectOutputStream(bos);
        oos.writeObject(result);
        oos.flush();
        outstream.flush();
        //connection.shutdownOutput();
    }
} // end RequestProcessor