/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.odmg;
import org.odmg.*;
import ojb.broker.*;
import ojb.broker.metadata.*;

/**
 * Class declaration
 *
 *
 * @author
 * @version %I%, %G%
 */
public class DatabaseImpl implements org.odmg.Database
{

	/**
	 * Creates new DatabaseImpl
	 */
	public DatabaseImpl()
	{
		nrm = NamedRootsMap.getInstance();
	}

	/**
	 * Open the named database with the specified access mode.
	 * Attempts to open a database when it has already been opened will result in
	 * the throwing of the exception <code>DatabaseOpenException</code>.
	 * A <code>DatabaseNotFoundException</code> is thrown if the database does not exist.
	 * Some implementations may throw additional exceptions that are also derived from
	 * <code>ODMGException</code>.
	 * @paramnameThe name of the database.
	 * @paramaccessModeThe access mode, which should be one of the static fields:
	 * <code>OPEN_READ_ONLY</code>, <code>OPEN_READ_WRITE</code>,
	 * or <code>OPEN_EXCLUSIVE</code>.
	 * @exceptionODMGExceptionThe database could not be opened.
	 */
	public void open(String name, int accessMode) throws ODMGException
	{
		try
		{
			// Get a Persistence Broker instance, to load repository
			broker = PersistenceBrokerFactory.createPersistenceBroker(name);
		}
		catch (Throwable ex)
		{
			throw new DatabaseNotFoundException("OJB can't open repository " + name + "\n"
												+ ex.getMessage());
		}
	}

	/**
	 * Close the database.
	 * After you have closed a database, further attempts to access objects in the
	 * database will cause the exception <code>DatabaseClosedException</code> to be thrown.
	 * Some implementations may throw additional exceptions that are also derived
	 * from <code>ODMGException</code>.
	 * @exception ODMGException Unable to close the database.
	 */
	public void close() throws ODMGException
	{
		OJB.getInstance().setCurrentDatabase(null);
	}

	/**
	 * Associate a name with an object and make it persistent.
	 * An object instance may be bound to more than one name.
	 * Binding a previously transient object to a name makes that object persistent.
	 * @param object The object to be named.
	 * @param name The name to be given to the object.
	 * @exception org.odmg.ObjectNameNotUniqueException
	 * If an attempt is made to bind a name to an object and that name is already bound
	 * to an object.
	 */
	public void bind(Object object, String name) throws ObjectNameNotUniqueException
	{
		// mark object as persistent first
		//this.makePersistent(object); //thma: this line produces errors in 6 TestCases!
		nrm.put(name, new Identity(object));
	}

	/**
	 * Lookup an object via its name.
	 * @paramnameThe name of an object.
	 * @return The object with that name.
	 * @exceptionObjectNameNotFoundExceptionThere is no object with the specified name.
	 * @seeObjectNameNotFoundException
	 * The lookup mechanism will search in the NamedRootsMap
	 * but it can also be used to lookup serialized OIDs: lookup(oid.serialize()).
	 * this is not specified by ODMG but can be useful in concunction with the
	 * String serializedOid = OJB.getObjectId(Object obj) method, which returns
	 * a serialized form of the Objects OID.
	 */
	public Object lookup(String name) throws ObjectNameNotFoundException
	{
		Identity oid = nrm.get(name);
		if (oid != null)
		{
			try
			{
				return broker.getObjectByIdentity(oid);
			}
			catch (PersistenceBrokerException ex)
			{
				throw new ObjectNameNotFoundException("Error in lookup of: " + name + ", " + ex.getMessage());
			}
		}
		else
		{
			throw new ObjectNameNotFoundException("Name not found: " + name);
		}
	}

	/**
	 * Disassociate a name with an object
	 * @param name The name of an object.
	 * @exception ObjectNameNotFoundException No object exists in the database with that name.
	 */
	public void unbind(String name) throws ObjectNameNotFoundException
	{
		nrm.unbind(name);
	}

	/**
	 * Make a transient object durable in the database.
	 * It must be executed in the context of an open transaction.
	 * If the transaction in which this method is executed commits,
	 * then the object is made durable.
	 * If the transaction aborts,
	 * then the makePersistent operation is considered not to have been executed,
	 * and the target object is again transient.
	 * ClassNotPersistenceCapableException is thrown if the implementation cannot make
	 * the object persistent because of the type of the object.
	 * @param	object	The object to make persistent.
	 */
	public void makePersistent(Object object)
	{
		try
		{
			// check whether object's class is persistence capable
			ClassDescriptor cld =
				DescriptorRepository.getInstance().getDescriptorFor(object.getClass());
			// if a ClassDescriptor was found, OJB knows how to persist the object
			TransactionImpl tx = (TransactionImpl) OJB.getInstance().currentTransaction();
			tx.register(object);
		}
		catch (ojb.broker.ClassNotPersistenceCapableException ex)
		{
			throw new org.odmg.ClassNotPersistenceCapableException(ex.getMessage());
		}
	}

	/**
	 * Deletes an object from the database.
	 * It must be executed in the context of an open transaction.
	 * If the object is not persistent, then ObjectNotPersistent is thrown.
	 * If the transaction in which this method is executed commits,
	 * then the object is removed from the database.
	 * If the transaction aborts,
	 * then the deletePersistent operation is considered not to have been executed,
	 * and the target object is again in the database.
	 * @param	object	The object to delete.
	 */
	public void deletePersistent(Object object)
	{
		TransactionImpl tx = (TransactionImpl) OJB.getInstance().currentTransaction();
		tx.register(object);
		tx.markDelete(object);
	}

	// the broker instance for persistence operations
	private PersistenceBroker broker = null;
	// the named roots map;
	private NamedRootsMap nrm;

	/**
	 * Insert the method's description here.
	 * Creation date: (11.01.2001 19:31:17)
	 * @return ojb.server.NamedRootsMap
	 */
	public NamedRootsMap getNamedRootsMap()
	{
		return nrm;
	}

	/**
	 * Insert the method's description here.
	 * Creation date: (26.01.2001 22:30:17)
	 * @return ojb.broker.PersistenceBroker
	 */
	public ojb.broker.PersistenceBroker getBroker()
	{
		return broker;
	}

}