/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.odmg;
import java.util.*;
import org.odmg.LockNotGrantedException;
import org.odmg.Transaction;
import ojb.odmg.states.*;
import ojb.broker.metadata.DescriptorRepository;
import ojb.odmg.lockstrategy.ReaderWriterLock;

import ojb.broker.ManageableCollection;
import ojb.broker.metadata.CollectionDescriptor;
import ojb.broker.ArrayIterator;
import ojb.broker.metadata.ClassDescriptor;
import ojb.broker.metadata.ObjectReferenceDescriptor;

/**
 * 
 * Implementation of Transaction for org.odmg.Transaction.
 * 
 * 
 * 
 * @author     Thomas Mahler & David Dixon-Peugh
 * 
 * @created    November 15, 2000
 * 
 */
public class TransactionImpl implements org.odmg.Transaction
{

	/**
	 * 
	 * the internal table containing all Objects "touched" by this tx and their
	 * 
	 * respective transactional state
	 * 
	 */
	private ObjectStateTable objectStateTable = null;

	/**
	 * 
	 * Internal table which provides mapping between threads and transactions.
	 * 
	 * This is required because a Thread can join a Transaction already in
	 * 
	 * progress.
	 * 
	 * 
	 * 
	 * If the thread joins a transaction, then "getTransaction()" should return
	 * 
	 * the apropriate one.  The only way we can ensure that is by keeping hold
	 * 
	 * of the txTable.
	 * 
	 */
	private TransactionTable txTable = null;

	/**
	 * 
	 * Used to store the DescriptorRepository.
	 * 
	 */
	private DescriptorRepository dr = DescriptorRepository.getInstance();

	/**
	 * 
	 * Determine the state.
	 * 
	 */
	private boolean _isOpen = false;

	/**
	 * 
	 * reference to the currently opened database
	 * 
	 */
	private DatabaseImpl curDB;

	/**
	 * 
	 * Determine whether the transaction is open or not. A transaction is open if
	 * 
	 * a call has been made to <code>begin</code> , but a subsequent call to
	 * 
	 * either <code>commit</code> or <code>abort</code> has not been made.
	 * 
	 * 
	 * 
	 * @return    True if the transaction is open, otherwise false.
	 * 
	 */
	public boolean isOpen()
	{
		return _isOpen;
	}

	/**
	 * 
	 * Attach the caller's thread to this <code>Transaction</code> and detach the
	 * 
	 * thread from any former <code>Transaction</code> the thread may have been
	 * 
	 * associated with.
	 * 
	 */
	public void join()
	{
		txTable.remove(Thread.currentThread());
		txTable.put(Thread.currentThread(), this);
	}

	/**
	 * 
	 * Upgrade the lock on the given object to the given lock mode. The call has
	 * 
	 * no effect if the object's current lock is already at or above that level of
	 * 
	 * lock mode.
	 * 
	 * 
	 * 
	 * @param  obj       object to acquire a lock on.
	 * 
	 * @param  lockMode  lock mode to acquire. The lock modes
	 * 
	 * are <code>READ</code> , <code>UPGRADE</code> , and <code>WRITE</code> .
	 * 
	 * @exception  LockNotGrantedException    Description of Exception
	 * 
	 */
	public void lock(Object obj, int lockMode) throws LockNotGrantedException
	{
		if (!_isOpen)
		{
			throw new org.odmg.TransactionNotInProgressException();
		}
		LockManager lm = LockManager.getInstance();
		ReaderWriterLock rwl = lm.getLock(obj);
		if (lockMode == Transaction.READ)
		{
			if (!rwl.read(this))
			{
				throw new LockNotGrantedException();
			}
		}
		if (lockMode == Transaction.WRITE)
		{
			if (!rwl.write(this))
			{
				throw new LockNotGrantedException();
			}
		}
		if (lockMode == Transaction.UPGRADE)
		{
			if (!rwl.upgrade(this))
			{
				throw new LockNotGrantedException();
			}
		}
		register(obj);
	}

	/**
	 * 
	 * Detach the caller's thread from this <code>Transaction</code> , but do not
	 * 
	 * attach the thread to another <code>Transaction</code> .
	 * 
	 */
	public void leave()
	{
		txTable.remove(Thread.currentThread());
	}

	/**
	 * 
	 * Do the Commits, but don't release the locks.
	 * 
	 * 
	 * 
	 * I don't know what we should do if we are in a checkpoint and
	 * 
	 * we need to abort.
	 * 
	 */
	private synchronized void doCommit() throws org.odmg.TransactionAbortedException
	{
		ObjectTransactionWrapper item;
		// Notify objects of impending commits.
		Enumeration enum = objectStateTable.elements();
		while (enum.hasMoreElements())
		{
			item = (ObjectTransactionWrapper) enum.nextElement();
			item.beforeCommit();
		}
		// Now perfom the real work
		objectStateTable.commit();
		// Now, we notify everything the commit is done.
		enum = objectStateTable.elements();
		while (enum.hasMoreElements())
		{
			item = (ObjectTransactionWrapper) enum.nextElement();
			item.afterCommit();
		}
	}

	/**
	 * Do the Aborts, but don't release the locks.
	 * 
	 * Do the aborts on the NamedRootsMap first, then
	 * abort the other stuff.
	 */
	private synchronized void doAbort()
	{
		ObjectTransactionWrapper item;
		Enumeration objectsToAbort = objectStateTable.elements();
		// Notify objects of impending aborts.
		while (objectsToAbort.hasMoreElements())
		{
			item = (ObjectTransactionWrapper) objectsToAbort.nextElement();
			item.beforeAbort();
		}
		// Now, we abort everything. . .
		objectStateTable.rollback();
		// Now, we notify everything the abort is done.
		objectsToAbort = objectStateTable.elements();
		while (objectsToAbort.hasMoreElements())
		{
			item = (ObjectTransactionWrapper) objectsToAbort.nextElement();
			item.afterAbort();
		}
	}

	/**
	 * Close a transaction and do all the cleanup associated with it.
	 */
	private synchronized void doClose()
	{
		Enumeration enum = objectStateTable.elements();
		while (enum.hasMoreElements())
		{
			removeLock(((ObjectTransactionWrapper) enum.nextElement()).getObject(), 
					   Transaction.WRITE);
		}
		_isOpen = false;
	}

	/**
	 * Commit the transaction, but reopen the transaction, retaining all locks.
	 * Calling <code>checkpoint</code> commits persistent object modifications
	 * made within the transaction since the last checkpoint to the database. The
	 * transaction retains all locks it held on those objects at the time the
	 * checkpoint was invoked.
	 */
	public void checkpoint()
	{
		if (!_isOpen)
		{
			throw new org.odmg.TransactionNotInProgressException();
		}
		try
		{
			doCommit();
		}
		catch (Exception e)
		{
			doAbort();
		}
	}

	/**
	 * 
	 * Upgrade the lock on the given object to the given lock mode. Method <code>
	 * 
	 * tryLock</code> is the same as <code>lock</code> except it returns a boolean
	 * 
	 * indicating whether the lock was granted instead of generating an exception.
	 * 
	 * 
	 * 
	 * @param  obj          Description of Parameter
	 * 
	 * @param  lockMode     Description of Parameter
	 * 
	 * @return              Description of the Returned Value
	 * 
	 * </code>, <code>UPGRADE</code> , and <code>WRITE</code> .
	 * 
	 * @returnTrue          if the lock has been acquired, otherwise false.
	 * 
	 */
	public boolean tryLock(Object obj, int lockMode)
	{
		if (!_isOpen)
		{
			throw new org.odmg.TransactionNotInProgressException();
		}
		try
		{
			lock(obj, lockMode);
			return true;
		}
		catch (LockNotGrantedException ex)
		{
			return false;
		}
	}

	/**
	 * 
	 * removeLock removes the transactions lock from the Object.
	 * 
	 */
	private void removeLock(Object obj, int lockType)
	{
		ReaderWriterLock rwl = LockManager.getInstance().getLock(obj);
		rwl.release(this);
	}

	/**
	 * Commit and close the transaction. Calling <code>commit</code> commits to
	 * the database all persistent object modifications within the transaction and
	 * releases any locks held by the transaction. A persistent object
	 * modification is an update of any field of an existing persistent object, or
	 * an update or creation of a new named object in the database. If a
	 * persistent object modification results in a reference from an existing
	 * persistent object to a transient object, the transient object is moved to
	 * the database, and all references to it updated accordingly. Note that the
	 * act of moving a transient object to the database may create still more
	 * persistent references to transient objects, so its referents must be
	 * examined and moved as well. This process continues until the database
	 * contains no references to transient objects, a condition that is guaranteed
	 * as part of transaction commit. Committing a transaction does not remove
	 * from memory transient objects created during the transaction.
	 * 
	 * The updateObjectList contains a list of all objects for which this transaction
	 * has write privledge to.  We need to update these objects.
	 */
	public void commit()
	{
		if (!_isOpen)
		{
			throw new org.odmg.TransactionNotInProgressException();
		}
		try
		{
			doCommit();
		}
		catch (org.odmg.TransactionAbortedException ex)
		{
			doAbort();
		}
		doClose();
	}

	/**
	 * 
	 * Abort and close the transaction. Calling abort abandons all persistent
	 * 
	 * object modifications and releases the associated locks. Aborting a
	 * 
	 * transaction does not restore the state of modified transient objects
	 * 
	 */
	public void abort()
	{
		if (!_isOpen)
		{
			throw new org.odmg.TransactionNotInProgressException();
		}
		doAbort();
		doClose();
	}

	/**
	 * Start a transaction. Calling <code>begin</code> multiple times on the same
	 * transaction object, without an intervening call to <code>commit</code> or
	 * <code>abort</code> , causes the exception <code>
	 * TransactionInProgressException</code> to be thrown on the second and
	 * subsequent calls. Operations executed before a transaction has been opened,
	 * or before reopening after a transaction is aborted or committed, have
	 * undefined results; these may throw a <code>
	 * TransactionNotInProgressException</code> exception.
	 */
	public void begin()
	{
		if (_isOpen)
		{
			throw new org.odmg.TransactionInProgressException();
		}
		objectStateTable = new ObjectStateTable();
		// mark tx as opened
		_isOpen = true;
	}

	/**
	 * 
	 * Creates new Transaction
	 * 
	 * @param theTxTable - Transaction table so we can manage this and
	 * 
	 * other transactions.
	 * 
	 */
	public TransactionImpl(TransactionTable theTxTable, DatabaseImpl theCurrentDB)
	{
		txTable = theTxTable;
		curDB = theCurrentDB;
	}



	/**
	 * 
	 * marks an object for deletion
	 * 
	 * 
	 * 
	 * @param  anObject Object to be marked
	 * 
	 */
	public void markDelete(Object anObject)
	{
		ObjectTransactionWrapper otw = objectStateTable.get(anObject);
		otw.setModificationState(otw.getModificationState().markDelete());
	}

/**
 * 
 * registers the object with this transaction recursively with all associations.
 * 
 */
public void register(Object newTxObject)
{
	if (!objectStateTable.contains(newTxObject))
	{
		// 1. register object itself
		objectStateTable.put(newTxObject, new ObjectTransactionWrapper(newTxObject));

		ClassDescriptor cld = null;
		// 2. register referenced objects
		try
		{
			cld =
				DescriptorRepository.getInstance().getDescriptorFor(newTxObject.getClass());
			Iterator i = cld.getObjectReferenceDescriptors().iterator();
			while (i.hasNext())
			{
				ObjectReferenceDescriptor rds = (ObjectReferenceDescriptor) i.next();
				Object refObj = rds.getPersistentField().get(newTxObject);
				register(refObj);
			}
		}
		catch (Exception ex)
		{
		}

		// 3. register objects in collection attributes
		try
		{

			Iterator i = cld.getCollectionDescriptors().iterator();
			while (i.hasNext())
			{
				CollectionDescriptor cds = (CollectionDescriptor) i.next();
				Object col = cds.getPersistentField().get(newTxObject);
				if (col != null)
				{
					Iterator colIterator;
					if (col instanceof ManageableCollection)
					{
						colIterator = ((ManageableCollection) col).ojbIterator();
					}
					else if (col instanceof Collection)
					{
						colIterator = ((Collection) col).iterator();
					}
					else if (col.getClass().isArray())
					{
						colIterator = new ArrayIterator(col);
					}
					else
					{
						throw new RuntimeException(
							col.getClass()
								+ " can not be managed by OJB, use Array, Collection or ManageableCollection instead !");
					}
					while (colIterator.hasNext())
					{
						register(colIterator.next());
					}
				}
			}
		}
		catch (Exception ex)
		{
		}
	}
	else
	{
	}
}

}