/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/
/*
 * ObJectRelationalBridge - Bridging Java Objects and Relational Databases
 * http://objectbridge.sourceforge.net
 * Copyright (C) 2000, 2001 Thomas Mahler, et al.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
package ojb.odmg.lockstrategy;

/**
 * Title:        ObJectRelationalBridge
 * Description:  Object Relational Mapping tool
 * Copyright:    Copyright (c) 2000
 * Company:
 * @author Thomas Mahler & David Dixon-Peugh
 * @version 1.0
 */
import java.util.*;
import org.odmg.Transaction;

/**
 * Class declaration
 * 
 * 
 * @author
 * @version %I%, %G%
 */
public class RWSerializableStrategy implements ReaderWriterLock
{
	private Set readers = new HashSet();
	private Transaction writer = null;

	/**
	 * Constructor declaration
	 * 
	 * 
	 * @see
	 */
	public RWSerializableStrategy() {}

	/**
	 * read -- Acquire a Read lock for the given Transaction.
	 * @param r - Transaction acquiring the lock.
	 * @return Whether lock was granted or not.
	 */
	public boolean read(Transaction r)
	{
		synchronized (this)
		{
			if (writer == r)
			{
				return true;    // If I'm the writer, I can read.
			}
			if (writer == null)
			{
				readers.add(r);
				return true;
			}
			return false;
		}
	}

	/**
	 * write -- Acquire a Write lock for the given transaction.
	 * @param w - Transaction acquiring the lock.
	 * @return wether the lock was granted or not.
	 */
	public boolean write(Transaction w)
	{
		synchronized (this)
		{
			if (writer == w)
			{
				return true;    // If I'm the writer, then I can write.
			}
			if ((writer == null) && (readers.size() == 0))
			{
				writer = w;
				return true;
			}
			return upgrade(w);
		}
	}

	/**
	 * upgrade -- Try to upgrade the lock to Write.
	 * @param u - Transaction trying to upgrade.
	 * @return Whether the upgrade was successful.
	 */
	public boolean upgrade(Transaction u)
	{
		if (writer == u)
		{
			return true;    // If I already have Write, then I've upgraded.
		}
		if (writer == null)
		{
			synchronized (this)
			{
				if (writer == null)
				{
					if (readers.size() == 1)
					{
						if (readers.contains(u))
						{
							readers.remove(u);
							writer = u;
							return true;
						}
					}
					else
					{
						if (readers.size() == 0)
						{
							writer = u;
							return true;
						}
					}
				}
			}
		}
		return false;
	}

	/**
	 * release -- Release the hold on the object.
	 * @param t Transaction to release lock.
	 */
	public boolean release(Transaction t)
	{
		synchronized (this)
		{
			if (writer == t)
			{
				writer = null;
				return true;
			}
			if (readers.contains(t))
			{
				readers.remove(t);
				return true;
			}
			return false;
		}
	}

	/**
	 * getLockType - Returns an "R" if it is a Read lock, "W"
	 * if it is a Write lock, or "0" (zero) if it isn't locked.
	 */
	public String getLockType(Transaction t)
	{
		if (readers.contains(t))
		{
			return "R";
		}
		if (t.equals(writer))
		{
			return "W";
		}
		return "0";
	}

	/**
	 * getStrategy -- Return ReaderWriterLock.RW_SERIALIZABLE
	 */
	public int getStrategy()
	{
		return ReaderWriterLock.RW_SERIALIZABLE;
	}

	/**
	 * checkRead -- Identify if Read access is granted.
	 */
	public boolean checkRead(Transaction tx)
	{
		if (readers.contains(tx))
		{
			return true;
		}
		if (tx.equals(writer))
		{
			return true;
		}
		return false;
	}

	/**
	 * checkWrite -- Determine if Write access is granted.
	 */
	public boolean checkWrite(Transaction tx)
	{
		return tx.equals(writer);
	}

}