package test.ojb.server;

/**
 * Title:        StateMachine Verifier
 * Description:  This project is to investigate using State Machine
 * Equivalence as a method of Software Testing.
 * Copyright:    Copyright (c) 2000
 * @author David Dixon-Peugh
 * @version 1.0
 */

import java.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.XMLReaderFactory;

public class StateMachineModel {
  private Map allStates = new HashMap();
  private Map spanningTree = new HashMap(); // STATE NAME x LIST (TRANSITIONS)
  private boolean spanOK = false;
  private ModelState startState;
  private Set actionSpace = new HashSet();
  private Set stateSpace = new HashSet();

  public StateMachineModel() {
  }

  /**
   * Load the state machine model from the given URI.
   */
  public void load( String URI ) {
    try {
      XMLReader parser =
        XMLReaderFactory.createXMLReader("org.apache.xerces.parsers.SAXParser");
      parser.setContentHandler(new FSMContentHandler( this ) );
      parser.parse(URI);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * Add a state to the State Machine.
   */
  public void addState( ModelState s ) {
    if (spanOK) spanOK = false;
    allStates.put( s.getID(), s );
    stateSpace.add( s );
    Iterator trans = s.getTransitions().iterator();

    while (trans.hasNext()) {
      ModelTransition t = (ModelTransition) trans.next();

      if (!actionSpace.contains( t.getName() )) {
        actionSpace.add( t.getName() );
      }
    }
  }

  /**
   * Returns a list of all states in the Finite State Machine.
   */
  public Set getStateSpace() {
    return stateSpace;
  }

  /**
   * Returns the set of available actions.
   */
  public Set getActionSpace() {
    return actionSpace;
  }

  /**
   * Returns a list of Transitions to get from Start to a
   * State.
   */
  public List getPath( ModelState s ) {
    if (!spanOK) calculateSpanningTree();
    return (List) spanningTree.get(s.getID());
  }

  private void addPath( ModelState state, List path ) {
    spanningTree.put( state.getID(), new ArrayList( path ) );
  }
  /**
   * Calculates a path for all states.
   */
  private void calculateSpanningTree( ) {
    List toDo = new ArrayList();
    int i = 0;

    toDo.add(getStartState());
    spanningTree.put( getStartState().getID(), new ArrayList( ) );

    while ( i < toDo.size() ) {
      ModelState s = (ModelState) toDo.get(i);
      Iterator trans = s.getTransitions().iterator();

      while (trans.hasNext()) {
        ModelTransition t = (ModelTransition) trans.next();

        if (!toDo.contains( getState( t.getTarget() ) )) {
          toDo.add(getState( t.getTarget() ));
          List newPath = new ArrayList( (List) spanningTree.get( s.getID() ) );
          newPath.add( t );
          spanningTree.put( t.getTarget(), newPath );
        }
      }
      i++;
    }
  }

  /**
   * Gets the start state
   */
  public ModelState getStartState() {
    return startState;
  }

  /**
   * Sets the start state
   */
  public void setStartState( ModelState newStart ) {
    startState = newStart;
  }

  /**
   * Gets the state associated with an ID.
   */
  public ModelState getState( String ID ) {
    return (ModelState) allStates.get(ID);
  }

  ////////////////////////////////////////////
  /**
   * ContentHandler for loading via SAX
   */
  private class FSMContentHandler implements ContentHandler {
    private StateMachineModel model;
    private String lastValue;
    private ModelState currentState;
    private boolean isStart = false;

    public FSMContentHandler(StateMachineModel SMM)
    {
      model = SMM;
    }

    public void setDocumentLocator( Locator locator ) { }
    public void startDocument() throws SAXException   { }
    public void endDocument() throws SAXException     { }
    public void startPrefixMapping( String prefix, String uri ) { }
    public void endPrefixMapping( String prefix ) { }
    public void processingInstruction( String target, String data )
      throws SAXException { }

    /**
     * startElement - The elements we want, are "state" and "transition"
     */
    public void startElement( String namespaceURI, String localName,
                              String rawName, Attributes attrs )
      throws SAXException
    {
      if (localName.equals("state")) {
        currentState = new ModelState();
        currentState.setName( attrs.getValue("name") );
        currentState.setID( attrs.getValue("id") );
        if (attrs.getValue("start") != null) {
          if (attrs.getValue("start").equals("Y")) {
            currentState.setStart(true);
          }
        }
      }

      if (localName.equals("trans")) {
        ModelTransition t = new ModelTransition();
        t.setName( attrs.getValue("name") );
        t.setTarget( attrs.getValue("target") );
        currentState.addTransition( t );
      }
    }

    /**
     * endElement - The elements we want are "state" and "transition"
     */
    public void endElement( String namespaceURI, String localName,
                            String rawName )
      throws SAXException
    {
      if (localName.equals("state")) {
        model.addState( currentState );
        if (currentState.isStart()) {
          model.setStartState( currentState );
        }
      }
    }

    /**
     * characters builds the lastValue string, just in case we need
     * it.  (Usually, you grab it in the endElement call.)
     */
    public void characters( char[] ch, int start, int end)
      throws SAXException
    {
      char []RC = new char[end - start];

      for (int i = start; i < end; i++) {
        RC[i-start] = ch[i];
      }

      lastValue = new String(RC);
    }

    public void ignorableWhitespace( char[] ch, int start, int end)
      throws SAXException { }
    public void skippedEntity(String name) throws SAXException { }

  }

  public static void showPath( StateMachineModel SMM, String state ) {
    List path;
    Iterator i;
    System.out.println(state + ": From start");

    if (SMM.getState( state ) == null) {
      System.out.println("  State Unreachable");
      return;
    }

    path = SMM.getPath(SMM.getState(state));

    i = path.iterator();

    while (i.hasNext()) {
      ModelTransition t = (ModelTransition) i.next();
      System.out.println("  to " + t.getTarget() + " by " + t.getName() );
    }
  }
  /**
   * Simple Test
   */
  public static void main(String args[]) {
    StateMachineModel SMM = new StateMachineModel();
    System.err.println( args[0] );
    SMM.load( args[0] );

    showPath( SMM, "00" );
    showPath( SMM, "0R" );
    showPath( SMM, "0W" );
    showPath( SMM, "R0" );
    showPath( SMM, "RR" );
    showPath( SMM, "RW" );
    showPath( SMM, "W0" );
    showPath( SMM, "WR" );
    showPath( SMM, "WW" );
  }
}